/*************************************************************************/ /*!
@File			innodpu_hdmi.c
@Title
@Copyright		Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
@Description
@License		Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/
#include <linux/platform_device.h>
#include <drm/drm_modes.h>
#include <drm/drm_edid.h>

#include "innodpu_hdmi.h"
#include "innodpu_hdmi_debugfs.h"
#include "innodpu_connector.h"

#ifdef CONFIG_DRM_INNO_AUDIO
#include "innoaudio_drv.h"
#endif

static int s_hdmi_hwi2c = false;
module_param(s_hdmi_hwi2c, int, 0600);
MODULE_PARM_DESC(s_hdmi_hwi2c, "hdmi hwi2c mode (default: false)");

static bool s_hdmi_support_75hz = true;
module_param(s_hdmi_support_75hz, bool, 0600);
MODULE_PARM_DESC(s_hdmi_support_75hz, "hdmi support 75hz (default: true)");

bool inno_hdmi_mode_is_push(struct hdmi_device_t *inno_hdmi, struct drm_display_mode *adjusted_mode)
{
	// TBD: RK??? not support
	if ( (adjusted_mode->htotal == inno_hdmi->native_mode.htotal) &&
		(adjusted_mode->vtotal == inno_hdmi->native_mode.vtotal) &&
		(adjusted_mode->clock == inno_hdmi->native_mode.clock) &&
		((adjusted_mode->hdisplay != inno_hdmi->native_mode.hdisplay)||
		(adjusted_mode->vdisplay != inno_hdmi->native_mode.vdisplay)) ) {
		return true;
	}

	return false;
}

bool inno_hdmi_native_mode_valid(struct hdmi_device_t *inno_hdmi)
{
	const struct drm_display_mode *mode = &inno_hdmi->native_mode;
	if ((mode->clock > 0) &&  ((mode->hdisplay > 0) && (mode->hdisplay < 8192)))
		return true;

	return false;
}

static bool inno_hdmi_encoder_mode_fixup(struct drm_encoder *encoder,
		const struct drm_display_mode *mode, struct drm_display_mode *adjusted_mode)
{
	struct hdmi_device_t *inno_hdmi
		= container_of(encoder, struct hdmi_device_t, encoder);

	if (inno_hdmi_native_mode_valid(inno_hdmi) && inno_hdmi_mode_is_push(inno_hdmi, adjusted_mode)) {
		drm_mode_copy(adjusted_mode, &inno_hdmi->native_mode);
	}

	return true;
}

static void inno_hdmi_encoder_mode_set(struct drm_encoder *encoder,
		struct drm_display_mode *mode, struct drm_display_mode *adjusted_mode)
{
	struct hdmi_device_t *inno_hdmi
		= container_of(encoder, struct hdmi_device_t, encoder);
	int dpu_id = 0;

	if (!encoder || !adjusted_mode) {
		return;
	}

	atomic64_set(&inno_hdmi->modesetting, 1);

	dpu_id = innodpu_get_dpuid_bycrtc(encoder->crtc);

	if (inno_hdmi->chip.encoder_modeset) {
		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI,"%s modeset: "DRM_MODE_FMT "\n",
			inno_hdmi->name, DRM_MODE_ARG(adjusted_mode));
		drm_mode_copy(inno_hdmi->chip.adjusted_mode, adjusted_mode);
		inno_hdmi->chip.encoder_modeset(&inno_hdmi->chip,dpu_id, inno_hdmi->chip.test_mode, adjusted_mode);
	}

	return;
}

static void inno_hdmi_encoder_mode_disable(struct drm_encoder *encoder)
{
	struct hdmi_device_t *inno_hdmi
		= container_of(encoder, struct hdmi_device_t, encoder);

	if (NULL == encoder) {
		return;
	}

#ifdef CONFIG_DRM_INNO_AUDIO
	if(inno_hdmi->ac && inno_hdmi->ac->report_jack){
		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI,
			"%s try report jack, conn_st:%d, has_audio:%d",
			inno_hdmi->name, inno_hdmi->ac->conn_st, inno_hdmi->ac->has_audio);

		inno_hdmi->ac->dpms = inno_hdmi->connector.dpms;
		inno_hdmi->ac->conn_st = inno_hdmi->chip.hpd_status_detect(&inno_hdmi->chip);
		inno_hdmi->ac->update_eld(inno_hdmi->ac, inno_hdmi->connector.eld, MAX_ELD_BYTES);
		inno_hdmi->ac->report_jack(inno_hdmi->ac);
	}
#endif

	if (encoder->crtc) {
		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI,
			"inno %s encoder mode disable\n",inno_hdmi->name );
	} else {
		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI,
			"inno %s encoder mode disable, crtc is null\n",inno_hdmi->name);
	}

	if (inno_hdmi->chip.encoder_disable)
		inno_hdmi->chip.encoder_disable(&inno_hdmi->chip);

	cancel_delayed_work_sync(&inno_hdmi->scramble_work);
}

static void inno_hdmi_encoder_mode_enable(struct drm_encoder *encoder)
{
	struct hdmi_device_t *inno_hdmi
		= container_of(encoder, struct hdmi_device_t, encoder);

	if (NULL == encoder) {
		goto out;
	}

	if (encoder->crtc) {
		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI,
			"inno %s encoder mode enable\n",inno_hdmi->name );
	} else {
		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI,
			"inno %s encoder mode enable, crtc is null\n",inno_hdmi->name);
	}
	if (inno_hdmi->chip.encoder_enable)
		inno_hdmi->chip.encoder_enable(&inno_hdmi->chip, encoder->crtc);

	if (inno_hdmi->chip.output_mode &&
		is_output_type_dvi(inno_hdmi->chip.output_mode))
		goto out;

#ifdef CONFIG_DRM_INNO_AUDIO
		if(inno_hdmi->ac && inno_hdmi->ac->report_jack){
			innodpu_info(inno_hdmi->dev, DPU_UT_HDMI,
					"try report jack, conn_st:%d, has_audio:%d",
					inno_hdmi->ac->conn_st, inno_hdmi->ac->has_audio);

			inno_hdmi->ac->dpms = inno_hdmi->connector.dpms;
			inno_hdmi->ac->conn_st = inno_hdmi->chip.hpd_status_detect(&inno_hdmi->chip);
			inno_hdmi->ac->update_eld(inno_hdmi->ac, \
							inno_hdmi->connector.eld, MAX_ELD_BYTES);
			inno_hdmi->ac->report_jack(inno_hdmi->ac);
		}
#endif

	innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, "enable to scramble work.\n");
	inno_hdmi->scramb_cnt = 0;
	inno_hdmi->scramb_gracecnt=30;
	cancel_delayed_work(&inno_hdmi->scramble_work);
	queue_delayed_work(inno_hdmi->scrambwq, &inno_hdmi->scramble_work, msecs_to_jiffies(10));

out:
	atomic64_set(&inno_hdmi->modesetting, 0);
}

static int inno_hdmi_encoder_destroy(struct drm_encoder *encoder)
{
	BUG_ON(encoder);

	/*
	 * Called  drm_encoder_cleanup
	 */
	if (encoder->funcs && encoder->funcs->destroy)
		encoder->funcs->destroy(encoder);

	return 0;
}

static const struct drm_encoder_funcs s_inno_hdmi_encoder_funcs = {
	.destroy = drm_encoder_cleanup,
};

static const struct drm_encoder_helper_funcs s_inno_hdmi_encoder_helper_funcs = {
	//.mode_valid = inno_hdmi_encoder_mode_valid,
	.mode_fixup = inno_hdmi_encoder_mode_fixup,
	.mode_set = inno_hdmi_encoder_mode_set,
	.disable = inno_hdmi_encoder_mode_disable,
	.enable = inno_hdmi_encoder_mode_enable,
};

static int inno_hdmi_encoder_create(struct drm_device *drm_dev,
		struct drm_encoder *encoder, unsigned int possible_crtc)
{
	int retcode = 0;

	drm_encoder_helper_add(encoder, &s_inno_hdmi_encoder_helper_funcs);
	retcode = drm_encoder_init(drm_dev, encoder,
		&s_inno_hdmi_encoder_funcs, DRM_MODE_ENCODER_TMDS, NULL);

	encoder->possible_crtcs = possible_crtc;

	return retcode;
}

static void inno_hdmi_connector_destroy_func(struct drm_connector *connector)
{
	if (!connector) {
		return;
	}

	drm_connector_unregister(connector);
	drm_connector_cleanup(connector);
}

static int inno_hdmi_get_edid(void *data, u8 * buf, u32 block, size_t len)
{
	struct hdmi_chip_t *chip = (struct hdmi_chip_t *)data;

	switch (block) {
	case 0:
		inno_memcpy(buf, chip->edid_buf, len);
		break;
	case 1:
		inno_memcpy(buf, chip->edid_buf + EDID_LENGTH, len);
		break;
	default:
		return -1;
	}

	return 0;
}

static int inno_hdmi_get_modes_by_autoedid(struct drm_connector *connector)
{
	int nums = 0;
	struct edid *edid = NULL;
	struct hdmi_device_t *inno_hdmi = NULL;

	if (!connector)
		return nums;

	inno_hdmi = container_of(connector, struct hdmi_device_t, connector);

	if (inno_hdmi->chip.hdmi_edid_read) {
		inno_hdmi->chip.hdmi_edid_read(&inno_hdmi->chip);
		edid = drm_do_get_edid(connector, inno_hdmi_get_edid, &inno_hdmi->chip);
	} else if (inno_hdmi->chip.connector_get_edid) {
		edid = inno_hdmi->chip.connector_get_edid(connector, &inno_hdmi->chip);
	}

	if (edid) {
		inno_hdmi->chip.max_clk = innodpu_conn_get_monitor_max_clk((u8 *)edid);
		if (inno_hdmi->chip.max_clk <= 0) {
			inno_hdmi->chip.max_clk = 605000;
			innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, "get hdmi max clock err, set it to 605MHz\n");
		}

		if (inno_hdmi->chip.replace_timing)
			innodpu_detailed_block_monitor_range_replace((u8 *)edid);

		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, "update_eld");
		nums = drm_add_edid_modes(connector, edid);

#if (DRM_VERSION <= KERNEL_VERSION(4, 16, 0))
		drm_edid_to_eld(connector, edid);
#endif
		inno_hdmi->chip.sink_is_hdmi = drm_detect_hdmi_monitor(edid);
		inno_hdmi->chip.sink_has_audio = drm_detect_monitor_audio(edid);

		/*if it is not hdmi sink for hdmi source, to force set dvi mode.*/
		if ((inno_hdmi->chip.output_mode->mode_connector_item == DRM_MODE_CONNECTOR_HDMIA) &&
			(!drm_detect_hdmi_monitor(edid))) {
			inno_hdmi->chip.output_mode->mode_connector_item = DRM_MODE_CONNECTOR_DVII;
		}

		drm_connector_update_edid_property(connector, edid);
#ifdef CONFIG_DRM_INNO_AUDIO
		if(inno_hdmi->ac){
			inno_hdmi->ac->has_audio = inno_hdmi->chip.sink_has_audio;
			innodpu_info(inno_hdmi->dev, DPU_UT_HDMI,
					"update audio param, conn_st:%d, has_audio:%d",
					inno_hdmi->ac->conn_st, inno_hdmi->ac->has_audio);
		}
#endif
		kfree(edid);
	}

	return nums;
}

static int inno_hdmi_get_modes_by_strpush(struct drm_connector *connector)
{
	int ret = -1;
	int nums = 0;
	struct hdmi_device_t *inno_hdmi = NULL;
	unsigned char str_push_edid_buf[INNOHDMI_EDID_BUF_LEN] = {0};

	if (!connector)
		return nums;

	inno_hdmi = container_of(connector, struct hdmi_device_t, connector);

	ret = hal_hdmi_edid_data(inno_hdmi->dev->parent, inno_hdmi->hdmi_id,
			str_push_edid_buf);
	if (ret < 0) {
		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, \
				"hal_hdmi_edid_data failed, return %d.\n", ret);
		return nums;
	}

	nums = innodpu_str_push_edid(str_push_edid_buf, connector);
	drm_mode_sort(&connector->probed_modes);

	return nums;
}

static int inno_hdmi_get_modes_by_userdefine(struct drm_connector *connector)
{
	int ret = -1;
	int nums = 0;
	struct edid *edid = NULL;
	struct hdmi_device_t *inno_hdmi = NULL;

	if (!connector)
		return nums;

	inno_hdmi = container_of(connector, struct hdmi_device_t, connector);

	ret = hal_hdmi_edid_data(inno_hdmi->dev->parent, inno_hdmi->hdmi_id,
			inno_hdmi->chip.edid_buf);
	if (ret < 0) {
		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, \
				"hal_hdmi_edid_data failed, return %d.\n", ret);
		return nums;
	}

	edid = drm_do_get_edid(connector, inno_hdmi_get_edid, &inno_hdmi->chip);
	if (edid) {
		inno_hdmi->chip.max_clk = innodpu_conn_get_monitor_max_clk((u8 *)edid);
		if (inno_hdmi->chip.max_clk <= 0) {
			inno_hdmi->chip.max_clk = 605000;
			innodpu_err(inno_hdmi->dev, "get hdmi max clock err, set it to 605MHz\n");
		}
		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI,"update eld");
		nums = drm_add_edid_modes(connector, edid);

#if (DRM_VERSION <= KERNEL_VERSION(4, 16, 0))
		drm_edid_to_eld(connector, edid);
#endif

		inno_hdmi->chip.sink_is_hdmi = drm_detect_hdmi_monitor(edid);
		inno_hdmi->chip.sink_has_audio = drm_detect_monitor_audio(edid);

		/*if it is not hdmi sink for hdmi source, to force set dvi mode.*/
		if ((inno_hdmi->chip.output_mode->mode_connector_item == DRM_MODE_CONNECTOR_HDMIA) &&
			(!drm_detect_hdmi_monitor(edid))) {
			inno_hdmi->chip.output_mode->mode_connector_item = DRM_MODE_CONNECTOR_DVII;
		}

		drm_connector_update_edid_property(connector, edid);
#ifdef CONFIG_DRM_INNO_AUDIO
		if(inno_hdmi->ac){
			inno_hdmi->ac->has_audio = inno_hdmi->chip.sink_has_audio;
			innodpu_info(inno_hdmi->dev, DPU_UT_HDMI,
					"update audio param, conn_st:%d, has_audio:%d",
					inno_hdmi->ac->conn_st, inno_hdmi->ac->has_audio);
		}
#endif
		kfree(edid);
	}

	return nums;
}

static bool inno_hdmi_native_mode_filter(struct drm_connector *connector,
		struct drm_display_mode *mode, bool priv_filter)
{
	struct hdmi_device_t *inno_hdmi
		= container_of(connector, struct hdmi_device_t, connector);
#if ((DRM_VERSION >= KERNEL_VERSION(4, 13, 0)))
	struct drm_encoder *encoder = &inno_hdmi->encoder;
#endif
	int ret = 0;

	// TBD: current not support 4k scaler
	if (mode && priv_filter && ((mode->hdisplay > INNODPU_COMBINE_WIDTH) ||
		(mode->vdisplay > INNODPU_COMBINE_HEIGHT)) )
		return true;

	if (mode && mode->clock > inno_hdmi->chip.max_clk)
		return true;

	if (connector->helper_private && connector->helper_private->mode_valid) {
		ret = connector->helper_private->mode_valid(&inno_hdmi->connector, mode);
		if (ret != MODE_OK) {
			return true;
		}
	}

#if ((DRM_VERSION >= KERNEL_VERSION(4, 13, 0)))
	if (encoder->helper_private && encoder->helper_private->mode_valid) {
		ret = encoder->helper_private->mode_valid(&inno_hdmi->encoder, mode);
		if (ret != MODE_OK) {
			return true;
		}
	}
#endif
	return false;
}

/**
 * inno_hdmi_connector_helper_get_modes - Get hdmi supports mode;
 *
 * Returns:
 * Number of supported modes
 *
 * 1. read edid and get modes
 * 2. push string
 * 3. replace timing
 * 4. get native mode
 * 5. add common modes
 */
static int inno_hdmi_connector_helper_get_modes(
		struct drm_connector *connector)
{
	int nums = 0;
	struct drm_display_mode *native_mode;
	struct hdmi_device_t *inno_hdmi
		= container_of(connector, struct hdmi_device_t, connector);

	switch (inno_hdmi->chip.hal_edid_mode) {
		case EDID_AUTO_READ:
			nums = inno_hdmi_get_modes_by_autoedid(connector);
			break;
		case EDID_STR_PUSH:
			nums = inno_hdmi_get_modes_by_autoedid(connector);
			if (nums <= 0) {
				innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, "zoom:can't read autoedid.\n");
			}
			nums += inno_hdmi_get_modes_by_strpush(connector);
			if (nums <= 0) {
				innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, "zoom:can't parse strpush.\n");
			}
			break;
		case EDID_USER_DEFINE:
			nums = inno_hdmi_get_modes_by_userdefine(connector);
			break;
		default:
			nums = 0;
			innodpu_warn(inno_hdmi->dev, "unkown hal_edid_mode(%d)!\n", inno_hdmi->chip.hal_edid_mode);
			break;
	}

	if (!nums) {
		inno_hdmi->chip.modes = innodpu_add_modes_without_edid(connector, NULL);
	} else {
		inno_hdmi->chip.modes = nums;
	}

	//drop repeat mode
	inno_hdmi->chip.modes -= innodpu_modes_drop_repeat(&inno_hdmi->connector);

	/*replace timing to cea standard*/
	if (inno_hdmi->chip.replace_timing)
		innodpu_modes_replace(&inno_hdmi->connector, NULL, NULL);

	// get native mode
	drm_mode_sort(&connector->probed_modes);
	native_mode = innodpu_get_native_mode(connector, inno_hdmi_native_mode_filter);

	// add common mode
	if (native_mode) {
		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, "%s native mode: "DRM_MODE_FMT ", status:%d\n",
			inno_hdmi->name, DRM_MODE_ARG(native_mode), native_mode->status);
		inno_memcpy(&inno_hdmi->native_mode, native_mode, sizeof(inno_hdmi->native_mode));
		drm_mode_set_crtcinfo(&inno_hdmi->native_mode, CRTC_INTERLACE_HALVE_V);
		inno_hdmi->native_mode.status = MODE_OK;
		inno_hdmi->chip.modes += innodpu_connector_add_common_modes(connector, &inno_hdmi->native_mode, false);
	}

	return inno_hdmi->chip.modes;
}


/**
 * inno_hdmi_connector_detect_ctx - check hdmi hpd status
 *
 * Returns:
 * drm_connector_status. (connector_status_connected or connector_status_disconnected)
 */
static int inno_hdmi_connector_detect_ctx(struct drm_connector *connector,
		struct drm_modeset_acquire_ctx *ctx, bool force)
{
	struct hdmi_device_t *inno_hdmi
		= container_of(connector, struct hdmi_device_t, connector);
	int conn_status = connector_status_unknown;

	if (inno_hdmi->chip.connector_detect) {
		conn_status = inno_hdmi->chip.connector_detect(&inno_hdmi->chip);
	} else {
		innodpu_warn(inno_hdmi->dev, "Not support %s chip detect handle !!!\n",
				inno_hdmi->name);
	}

	innodpu_info(inno_hdmi->dev, DPU_UT_HDMI,
		"%s connector detect ctx: %s\n", inno_hdmi->name,
		inno_drm_get_connector_status_name(conn_status));

	return conn_status;
}

static enum drm_connector_status inno_hdmi_detect(struct drm_connector *connector, bool force)
{
    return inno_hdmi_connector_detect_ctx(connector, NULL, force);
}

static enum drm_mode_status inno_hdmi_connector_helper_mode_valid(
		struct drm_connector *connector, struct drm_display_mode *mode)
{
	struct hdmi_device_t *inno_hdmi
		= container_of(connector, struct hdmi_device_t, connector);
	int status = MODE_OK;

	if (mode->flags & DRM_MODE_FLAG_DBLSCAN) {
		return MODE_NO_DBLESCAN;
	}

	if (mode->flags & DRM_MODE_FLAG_INTERLACE) {
		return MODE_NO_INTERLACE;
	}

	if (!s_hdmi_support_75hz && drm_mode_vrefresh(mode) == 75) {
		return MODE_NOMODE;
	}

//#define HDMI_MODE_NEED_BYPASS
#ifdef  HDMI_MODE_NEED_BYPASS
	{
		struct mfc_monitor_info monitor = MFC_MONITOR_INFO_ITEM("LEN", "*");
		struct resolution_info  resolution[] = {
			RESOLUTION_INFO_ITEM(1920, 1080, 75, -1),
			RESOLUTION_INFO_ITEM(2560, 1440, 60, -1),
		};
		struct resolution_info  *match_resolution = NULL;

		if (innodpu_is_mfc_monitor_match((unsigned char *)inno_hdmi->chip.edid_buf, &monitor)) {
			match_resolution = innodpu_resolution_match(mode, resolution, INNO_ARRAY_SIZE(resolution));
			if (match_resolution) {
				innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, \
						"match resolution[%dx%d@%d %dKHz], drop it for monitor(LEN/*)\n",
						match_resolution->hdisplay, match_resolution->vdisplay, match_resolution->vrefresh);
				return MODE_NOMODE;
			}
		}
	}
#endif

	if (mode->hdisplay % 16) {
		return MODE_BAD_HVALUE;
	}

	if (inno_hdmi->chip.connector_mode_valid) {
		status = inno_hdmi->chip.connector_mode_valid(&inno_hdmi->chip, mode);
		if (status != MODE_OK)
			return status;
	}

	return MODE_OK;
}

#if ((DRM_VERSION >= KERNEL_VERSION(4, 10, 0)))
static void inno_hdmi_connector_print_state(struct drm_printer *p,
					     const struct drm_connector_state *state)
{

}
#endif

static int inno_hdmi_connector_late_register(struct drm_connector *connector)
{
	struct hdmi_device_t * inno_hdmi = container_of(connector, struct hdmi_device_t, connector);
	int ret = 0;
	struct i2c_adapter *adapter = NULL;
	struct device *adapter_dev  = NULL;

	BUG_ON(!(connector->kdev));

	if (inno_hdmi->i2c) {
		adapter = &inno_hdmi->i2c->adapter;
		adapter_dev  = &adapter->dev;

		ret = sysfs_create_link(&connector->kdev->kobj,
				&adapter_dev->kobj,
				adapter_dev->kobj.name);
		if (ret) {
			innodpu_err(inno_hdmi->dev, "%s sysfs link create failed-%d\n",
				  adapter_dev->kobj.name, ret);
		}
	}

	ret = inno_hdmi_custom_debugfs_create(connector->debugfs_entry, &inno_hdmi->chip);
	if (ret) {
		innodpu_err(inno_hdmi->dev, "%s custom debugfs create failed-%d\n",
			  inno_hdmi->name, ret);
	}

	hal_dev_enable_irq(inno_hdmi->dev->parent, inno_hdmi->chip.hal_module);

	return 0;
}

static void inno_hdmi_connector_early_unregister(struct drm_connector *connector)
{
	struct hdmi_device_t *inno_hdmi
		= container_of(connector, struct hdmi_device_t, connector);
	struct i2c_adapter *adapter = NULL;
	struct device *adapter_dev  = NULL;

	BUG_ON(!(connector->kdev));

	hal_dev_disable_irq(inno_hdmi->dev->parent, inno_hdmi->chip.hal_module);

	inno_hdmi_custom_debugfs_remove(connector->debugfs_entry, &inno_hdmi->chip);

	if (inno_hdmi->i2c) {
		adapter = &inno_hdmi->i2c->adapter;
		adapter_dev  = &adapter->dev;
		sysfs_remove_link(&connector->kdev->kobj, adapter_dev->kobj.name);
	}
}

static struct drm_connector_helper_funcs s_inno_hdmi_connector_helper_funcs = {
	.get_modes  = inno_hdmi_connector_helper_get_modes,
	.mode_valid = inno_hdmi_connector_helper_mode_valid,
#if ((DRM_VERSION >= KERNEL_VERSION(4, 12, 0)))
	.detect_ctx = inno_hdmi_connector_detect_ctx,
#endif
};


static const struct drm_connector_funcs s_inno_hdmi_connector_funcs = {
	.dpms       = drm_helper_connector_dpms,
	.fill_modes = drm_helper_probe_single_connector_modes,
	.destroy = inno_hdmi_connector_destroy_func,
#if ((DRM_VERSION >= KERNEL_VERSION(4, 10, 0)))
	.atomic_print_state = inno_hdmi_connector_print_state,
#endif
	.reset = drm_atomic_helper_connector_reset,
	.atomic_duplicate_state = drm_atomic_helper_connector_duplicate_state,
	.atomic_destroy_state = drm_atomic_helper_connector_destroy_state,
	.late_register = inno_hdmi_connector_late_register,
	.early_unregister = inno_hdmi_connector_early_unregister,
	.detect = inno_hdmi_detect,
};

static int inno_hdmi_connector_create(struct drm_device *drm_dev,
				struct drm_connector *connector)
{
	int retcode = 0;
	struct hdmi_device_t * hdmi_dev = container_of(connector, struct hdmi_device_t, connector);

	BUG_ON(!drm_dev);
	BUG_ON(!connector);

	drm_connector_helper_add(connector, &s_inno_hdmi_connector_helper_funcs);

	if (hdmi_dev->chip.output_mode) {
		retcode = drm_connector_init(drm_dev, connector, &s_inno_hdmi_connector_funcs,
				hdmi_dev->chip.output_mode->mode_connector_item);
	} else {
		retcode = drm_connector_init(drm_dev, connector, &s_inno_hdmi_connector_funcs,
			DRM_MODE_CONNECTOR_HDMIA);
	}
	connector->polled = DRM_CONNECTOR_POLL_HPD;
	connector->dpms = DRM_MODE_DPMS_OFF;
	connector->interlace_allowed = false;
	connector->doublescan_allowed = false;
	connector->display_info.subpixel_order = SubPixelHorizontalRGB;

	return retcode;
}

static int inno_hdmi_connector_destroy(struct drm_connector *connector)
{
	struct drm_mode_object *conn_obj = &connector->base;

	while (kref_read(&conn_obj->refcount) > 0)
		drm_connector_put(connector); // will be called drm_connector_free

	return 0;
}


static int inno_hdmi_connector_attach_encoder(
			struct hdmi_device_t *inno_hdmi)
{
	int retcode = 0;

	retcode = inno_hdmi_encoder_create(inno_hdmi->drm_dev, &inno_hdmi->encoder,
		inno_hdmi->chip.possible_crtc);
	if (retcode) {
		innodpu_err(inno_hdmi->dev, "%s encoder create failed-%d\n",
				  inno_hdmi->name, retcode);
		return retcode;
	}

	retcode = inno_hdmi_connector_create(inno_hdmi->drm_dev, &inno_hdmi->connector);
	if (retcode) {
		innodpu_err(inno_hdmi->dev, "%s connector create failed-%d\n",
			inno_hdmi->name, retcode);
		goto err_connector_create;
	}

	// only set connector->possible_encoder,
	retcode = drm_connector_attach_encoder(&inno_hdmi->connector, &inno_hdmi->encoder);
	if (retcode) {
		innodpu_err(inno_hdmi->dev, "%s connector attach encoder failed-%d\n",
			inno_hdmi->name, retcode);
		goto error_attach_failed;
	}

	innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, "%s attach encoder-%d connector-%d\n",
		inno_hdmi->name, inno_hdmi->encoder.base.id, inno_hdmi->connector.base.id);

	return retcode;

error_attach_failed:
	inno_hdmi_connector_destroy(&inno_hdmi->connector);
err_connector_create:
	inno_hdmi_encoder_destroy(&inno_hdmi->encoder);

	return retcode;
}

static void inno_hdmi_cleanup_connector_encoder(
			struct hdmi_device_t *inno_hdmi)
{
	/*
	 * After the hdmi reference count is zeroed, drm_connector_free will be called
	 * encoder autorelease function without reference counting
	 */
	inno_hdmi_connector_destroy(&inno_hdmi->connector);
	inno_hdmi_encoder_destroy(&inno_hdmi->encoder);
}

static int inno_hdmi_edid_fixup(struct drm_connector *connector)
{
	int ret = -1;
	struct edid *edid = NULL;
	struct hdmi_device_t *inno_hdmi
		= container_of(connector, struct hdmi_device_t, connector);

	inno_hdmi = container_of(connector, struct hdmi_device_t, connector);
	switch (inno_hdmi->chip.hal_edid_mode) {
	case EDID_AUTO_READ:
		if (inno_hdmi->chip.hdmi_edid_read) {
			inno_hdmi->chip.hdmi_edid_read(&inno_hdmi->chip);
			edid = drm_do_get_edid(connector, inno_hdmi_get_edid, &inno_hdmi->chip);
		} else if (inno_hdmi->chip.connector_get_edid) {
			edid = inno_hdmi->chip.connector_get_edid(connector, &inno_hdmi->chip);
		}
		break;
	case EDID_USER_DEFINE:
		ret = hal_hdmi_edid_data(inno_hdmi->dev->parent, inno_hdmi->hdmi_id,
				inno_hdmi->chip.edid_buf);
		if (ret < 0) {
			innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, \
					"hal_hdmi_edid_data failed, return %d.\n", ret);
			return -1;
		}
		edid = drm_do_get_edid(connector, inno_hdmi_get_edid, &inno_hdmi->chip);
		break;
	default:
		return -1;
	}

	if (edid) {
#if (DRM_VERSION <= KERNEL_VERSION(4, 16, 0))
		drm_edid_to_eld(connector, edid);
#endif

		inno_hdmi->chip.sink_is_hdmi = drm_detect_hdmi_monitor(edid);
		inno_hdmi->chip.sink_has_audio = drm_detect_monitor_audio(edid);
		drm_connector_update_edid_property(connector, edid);

		kfree(edid);
		return 0;
	}

	return -1;
}

static void inno_hdmi_hpdin_fixup(struct drm_connector *connector)
{
	struct hdmi_device_t *inno_hdmi
		= container_of(connector, struct hdmi_device_t, connector);
	struct drm_encoder *encoder = &inno_hdmi->encoder;
	struct drm_crtc *crtc = encoder->crtc;

	if (crtc && crtc->state && crtc->state->active && !atomic64_read(&inno_hdmi->modesetting)) {
		struct drm_display_mode *mode = NULL;

		mode = inno_hdmi->chip.adjusted_mode;

		if (IS_ERR_OR_NULL(mode)) {
			innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, "recover-mode is NULL\n");
			return;
		}

		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, "%s start recover-%dx%d@%d\n",
				inno_hdmi->name, mode->hdisplay, mode->vdisplay, drm_mode_vrefresh(mode));

		/*
		 * TODO:
		 * if edid is null, we try to update it.
		 * but there is not a good way, we will find out why we didn't get
		 * the edid and solve it, rather than the inno_hdmi_edid_fixup function.
		 * */

		if (!((inno_drm_edid_block_checksum(inno_hdmi->chip.edid_buf) == 0) &&
			(inno_drm_edid_header_is_valid(inno_hdmi->chip.edid_buf) == 8))) {
			innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, "Try to fixup edid\n");
			inno_hdmi_edid_fixup(connector);
		}
		inno_hdmi_encoder_mode_set(encoder, NULL, mode);
#ifdef CONFIG_DRM_INNO_AUDIO
		if(inno_hdmi->ac){
			inno_hdmi->ac->has_audio = inno_hdmi->chip.sink_has_audio;
			innodpu_info(inno_hdmi->dev, DPU_UT_HDMI,
							"conn_st:%d, has_audio:%d",
							inno_hdmi->ac->conn_st, inno_hdmi->ac->has_audio);
		}
#endif
		inno_hdmi_encoder_mode_enable(encoder);
	}
}


static int inno_hdmi_hpdout_clean(struct drm_connector *connector)
{
	struct hdmi_device_t *inno_hdmi
		= container_of(connector, struct hdmi_device_t, connector);

	inno_hdmi->chip.sink_is_hdmi   = false;
	inno_hdmi->chip.sink_has_audio = false;
	innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, "edid clear ok\n");

#ifdef CONFIG_DRM_INNO_AUDIO
	if(inno_hdmi->ac){
		inno_hdmi->ac->has_audio = 0;
		innodpu_connector_clear_eld(&inno_hdmi->connector);
		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI,"conn_st:%d, has_audio:%d",
						inno_hdmi->ac->conn_st, inno_hdmi->ac->has_audio);
	}
#endif
	drm_connector_update_edid_property(connector, NULL);
	inno_memset(&inno_hdmi->native_mode, 0, sizeof(inno_hdmi->native_mode));
	inno_memset(inno_hdmi->chip.edid_buf, 0, INNOHDMI_EDID_BUF_LEN);

	drm_connector_update_edid_property(connector, NULL);
	cancel_delayed_work_sync(&inno_hdmi->scramble_work);

	return 0;
}

/*
 * If the ddc detection is more than N times, the HPD signal is not elevated,
 * which means that the display should be pulled out.
 *
 */
#define INNOHDMI_HPDOUT_THRESHOLD (5U)
static void __attribute__ ((unused)) inno_hdmi_hotplug_work(struct work_struct *work)
{
	struct hdmi_device_t *inno_hdmi =
		container_of(work, struct hdmi_device_t, hotplug_work.work);

	int hpd_status, ddc_status;

	/* when hpd-pin is high, should to report hpd-in. */
	if (inno_hdmi->chip.hpd_status_detect) {
		hpd_status = inno_hdmi->chip.hpd_status_detect(&inno_hdmi->chip);
		if (hpd_status) {
			inno_hdmi->hpdout_cnt = 0;
			inno_hdmi_hpdin_fixup(&inno_hdmi->connector);
			drm_helper_hpd_irq_event(inno_hdmi->drm_dev);
			return;
		}
	/* in order to support old way.*/
	} else if (inno_hdmi->chip.connector_detect) {
		if (connector_status_connected == \
			inno_hdmi->chip.connector_detect(&inno_hdmi->chip)) {
			inno_hdmi->hpdout_cnt = 0;
			drm_helper_hpd_irq_event(inno_hdmi->drm_dev);
			return;
		}
	}

	/* when hpd low happend, should to read ddc status,
	 * when ddc is okay, to ignore it.
	 * because it could be connected, the user to close hdmi signel out.
	 * when ddc is failed, to report hpd-out.
	 * */
	if (inno_hdmi->chip.ddc_status_detect && \
			!is_flags_skip_ddc_detect(inno_hdmi->chip.output_mode)) {
		ddc_status = inno_hdmi->chip.ddc_status_detect(&inno_hdmi->chip);
		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, \
				"loop ddc_detect:%s, hpdout_cnt:%d.\n",\
				ddc_status ? "okay" : "failed", inno_hdmi->hpdout_cnt);
		if (ddc_status && inno_hdmi->hpdout_cnt < INNOHDMI_HPDOUT_THRESHOLD) {
			inno_hdmi->hpdout_cnt ++;
			cancel_delayed_work(&inno_hdmi->hotplug_work);
			queue_delayed_work(inno_hdmi->hpdwq, &inno_hdmi->hotplug_work, \
					msecs_to_jiffies(1000));
			return;
		} else {
			inno_hdmi->hpdout_cnt = 0;
			inno_hdmi_hpdout_clean(&inno_hdmi->connector);
			drm_helper_hpd_irq_event(inno_hdmi->drm_dev);
			return;
		}
	} else {
		//force to report hpd-out
		inno_hdmi_hpdout_clean(&inno_hdmi->connector);
		drm_helper_hpd_irq_event(inno_hdmi->drm_dev);
		return;
	}
}

static void __attribute__ ((unused)) inno_hdmi_scramble_work(struct work_struct *work)
{
	struct hdmi_device_t *inno_hdmi =
		container_of(work, struct hdmi_device_t, scramble_work.work);
	int ret = 0;

	if (inno_hdmi->chip.try_scramble) {
		ret = inno_hdmi->chip.try_scramble(&inno_hdmi->chip, 0x3);
	}

	if (ret < 0) {
		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, "scramble again(cnt:%d).\n", inno_hdmi->scramb_cnt);
		inno_hdmi->scramb_cnt++;
		if (inno_hdmi->scramb_cnt < 30) {
			cancel_delayed_work(&inno_hdmi->scramble_work);
			queue_delayed_work(inno_hdmi->scrambwq, &inno_hdmi->scramble_work, msecs_to_jiffies(200));
		}
	} else if (ret > 0) {
		innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, "unscramble OK(cnt:%d).\n", inno_hdmi->scramb_cnt);
	} else {
		if (inno_hdmi->scramb_gracecnt) {
			inno_hdmi->scramb_gracecnt--;
			innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, "scramble grace(cnt:%d).\n", inno_hdmi->scramb_gracecnt);
			cancel_delayed_work(&inno_hdmi->scramble_work);
			queue_delayed_work(inno_hdmi->scrambwq, &inno_hdmi->scramble_work, msecs_to_jiffies(200));
		} else {
			innodpu_info(inno_hdmi->dev, DPU_UT_HDMI, "scramble OK(cnt:%d).\n", inno_hdmi->scramb_cnt);
		}
	}
}

static void inno_hdmi_irq_handle(void *data)
{
	unsigned int irq_event = 0;
	struct hdmi_device_t *inno_hdmi = (struct hdmi_device_t *)data;
	unsigned int delay_ms   = 500;
	int hpd_status = 0;

	BUG_ON(!inno_hdmi);

	if (inno_hdmi->chip.irq_handle) {
		irq_event = inno_hdmi->chip.irq_handle(&inno_hdmi->chip);
	} else {
		innodpu_warn(inno_hdmi->dev, "Not support %s chip irq handle !!!\n",
				  inno_hdmi->name);
		return;
	}

	hpd_status = inno_hdmi->chip.hpd_status_detect(&inno_hdmi->chip);

#ifdef CONFIG_DRM_INNO_AUDIO
	if (inno_hdmi->ac) {
		inno_hdmi->ac->conn_st = hpd_status;
	}
#endif

	/* when hpd is low and delay longer flags is set, we to use new delay_ms .*/
	if (!hpd_status && is_flags_wait_longer_hpd(inno_hdmi->chip.output_mode)) {
		delay_ms = inno_hdmi->chip.output_mode->data;
	}

	if (irq_event & HDMI_IRQ_HPD_MASK) {
		cancel_delayed_work(&inno_hdmi->hotplug_work);
		queue_delayed_work(inno_hdmi->hpdwq, &inno_hdmi->hotplug_work, msecs_to_jiffies(delay_ms));
		//innodpu_info(inno_hdmi->dev, DPU_UT_HDMI,"%s hot plug irq %#x\n", inno_hdmi->name);
	}

	return;
}

void inno_hdmi_hw_irq_enable(struct hdmi_device_t *inno_hdmi, unsigned int flags)
{
	BUG_ON(!inno_hdmi);
	if (inno_hdmi->chip.irq_enable)
		return inno_hdmi->chip.irq_enable(&inno_hdmi->chip, flags);
}

void inno_hdmi_hw_irq_disable(struct hdmi_device_t *inno_hdmi, unsigned int flags)
{
	BUG_ON(!inno_hdmi);
	if (inno_hdmi->chip.irq_disable)
		return inno_hdmi->chip.irq_disable(&inno_hdmi->chip, flags);
}

static int innohdmi_get_clock(void *data)
{
	struct hdmi_device_i2c_t *i2c = (struct hdmi_device_i2c_t *)data;
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&i2c->hdmi_dev->chip;

	if (chip && chip->chipi2c && chip->chipi2c->getscl) {
		return chip->chipi2c->getscl(chip);
	}

	return -1;
}

static int innohdmi_get_data(void *data)
{
	struct hdmi_device_i2c_t *i2c = (struct hdmi_device_i2c_t *)data;
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&i2c->hdmi_dev->chip;

	if (chip && chip->chipi2c && chip->chipi2c->getsda) {
		return chip->chipi2c->getsda(chip);
	}

	return -1;
}

static void innohdmi_set_clock(void *data, int state)
{
	struct hdmi_device_i2c_t *i2c = (struct hdmi_device_i2c_t *)data;
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&i2c->hdmi_dev->chip;

	if (chip && chip->chipi2c && chip->chipi2c->setscl) {
		chip->chipi2c->setscl(chip, state);
	}

	return;
}

static void innohdmi_set_data(void *data, int state)
{
	struct hdmi_device_i2c_t *i2c = (struct hdmi_device_i2c_t *)data;
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&i2c->hdmi_dev->chip;

	if (chip && chip->chipi2c && chip->chipi2c->setsda) {
		chip->chipi2c->setsda(chip, state);
	}

	return;
}


#define I2C_RISEFALL_TIME 20
static int innohdmi_pre_xfer(struct i2c_adapter *adapter)
{
	int ret;
	struct hdmi_device_i2c_t *i2c = container_of(adapter,
					       struct hdmi_device_i2c_t,
					       adapter);
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&i2c->hdmi_dev->chip;

	if (chip && chip->chipi2c && chip->chipi2c->pre_xfer) {
		ret = chip->chipi2c->pre_xfer(chip);
		if (ret)
			return ret;
	}

	innohdmi_set_data(i2c, 1);
	innohdmi_set_clock(i2c, 1);

	inno_udelay(I2C_RISEFALL_TIME);
	return 0;
}

static void innohdmi_post_xfer(struct i2c_adapter *adapter)
{
	struct hdmi_device_i2c_t *i2c = container_of(adapter,
					       struct hdmi_device_i2c_t,
					       adapter);
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&i2c->hdmi_dev->chip;

	if (chip && chip->chipi2c && chip->chipi2c->post_xfer) {
		chip->chipi2c->post_xfer(chip);
	}

	innohdmi_set_data(i2c, 1);
	innohdmi_set_clock(i2c, 1);

	inno_udelay(I2C_RISEFALL_TIME);
}

static int innohdmi_i2c_bit_xfer(struct i2c_adapter *adapter,
		struct i2c_msg *msgs, int num)
{
	extern const struct i2c_algorithm i2c_bit_algo;
	return i2c_bit_algo.master_xfer(adapter, msgs, num);
}

#define I2C_M_RD 0x0001
static int innohdmi_i2c_hw_xfer(struct i2c_adapter *adapter, struct i2c_msg *msgs, int num)
{
	struct hdmi_device_i2c_t *i2c = container_of(adapter,
					       struct hdmi_device_i2c_t,
					       adapter);
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&i2c->hdmi_dev->chip;
	int i, ret = -1;
	u16 addr, len;
	u8 *buf;

	if (chip && chip->chipi2c && chip->chipi2c->set_ddc_clk) {
		chip->chipi2c->set_ddc_clk(chip);
	}

	if (!chip->dev)
		return ret;

	for (i = 0; i < num; i++) {
		innodpu_info(chip->dev, DPU_UT_HDMI, "hwi2c xfer: num: %d/%d, addr:0x%02x, len: %d, flags: 0x%04x.\n",
				i + 1, num, msgs[i].addr, msgs[i].len, msgs[i].flags);

		addr = (msgs[i].flags & I2C_M_RD) ? ((msgs[i].addr<<1) | 0x1) : (msgs[i].addr<<1);
		buf = msgs[i].buf;
		len = msgs[i].len;

		if (chip && chip->chipi2c && chip->chipi2c->hwi2c_transfer) {
			ret = chip->chipi2c->hwi2c_transfer(chip, addr, buf, len);
			if (ret < 0)
				break;
		}
	}

	if (!ret) {
		ret = num;
	}

	return ret;
}

static int innohdmi_i2c_xfer(struct i2c_adapter *adapter,
		struct i2c_msg *msgs, int num)
{
	int ret = -1;
	struct hdmi_device_i2c_t *i2c = container_of(adapter,
					       struct hdmi_device_i2c_t,
					       adapter);
	struct hdmi_device_t *hdmi_dev = i2c->hdmi_dev;
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&hdmi_dev->chip;
	int (*xfer)(struct i2c_adapter *, struct i2c_msg *, int) = NULL;

	if (s_hdmi_hwi2c) {
		xfer = innohdmi_i2c_hw_xfer;
	} else {
		xfer = innohdmi_i2c_bit_xfer;
	}

	inno_mutex_lock(chip->chipi2c->mutex);
	ret = xfer(adapter, msgs, num);
	inno_mutex_unlock(chip->chipi2c->mutex);

	return ret;
}

static u32 innohdmi_i2c_functionality(struct i2c_adapter *adapter)
{
	/*
	struct hdmi_device_i2c_t *i2c = container_of(adapter,
					       struct hdmi_device_i2c_t,
					       adapter);

	return i2c->algorithm->functionality(adap, msgs, num);
	*/

	return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL;
}

static const struct i2c_algorithm innohdmi_algorithm = {
	.master_xfer = innohdmi_i2c_xfer,
	.functionality = innohdmi_i2c_functionality,
};

int innohdmi_i2c_init(struct hdmi_device_t *hdmi_dev)
{
	struct hdmi_device_i2c_t *i2c = NULL;
	struct i2c_adapter *adapter = NULL;
	struct i2c_algo_bit_data *bit_algo = NULL;

	i2c = kzalloc(sizeof(struct hdmi_device_i2c_t), GFP_KERNEL);
	if (!i2c)
		return -ENOMEM;

	adapter = &i2c->adapter;
	bit_algo = &i2c->bit_algo;

	snprintf(adapter->name, sizeof(adapter->name), \
			"hdmi%d-i2c", hdmi_dev->hdmi_id);
	adapter->algo = &innohdmi_algorithm;
	adapter->class = I2C_CLASS_DDC;
	adapter->dev.parent = hdmi_dev->dev;
	adapter->nr = -1;
	adapter->owner = THIS_MODULE;

	bit_algo->getscl = innohdmi_get_clock;
	bit_algo->getsda = innohdmi_get_data;
	bit_algo->setscl = innohdmi_set_clock;
	bit_algo->setsda = innohdmi_set_data;
	bit_algo->pre_xfer = innohdmi_pre_xfer;
	bit_algo->post_xfer = innohdmi_post_xfer;
	bit_algo->udelay = I2C_RISEFALL_TIME;
	bit_algo->timeout = usecs_to_jiffies(2200);
	bit_algo->data = i2c;
	//bit_algo->can_do_atomic = true;
	adapter->retries = 3;
	adapter->algo_data = bit_algo;

	if (i2c_add_adapter(adapter)) {
		goto err;
	}

	//i2c->hw_ddc_ops  = ddc_ops;
	i2c->hdmi_dev = hdmi_dev;
	hdmi_dev->i2c = i2c;

	return 0;

err:
	kfree(i2c);
	return -EINVAL;
}


static int inno_hdmi_hw_init(struct hdmi_device_t *inno_hdmi)
{
	int retcode = 0;

	if (inno_hdmi->chip.hw_init)
		return inno_hdmi->chip.hw_init(&inno_hdmi->chip);

	return retcode;
}

static void inno_hdmi_hw_fini(struct hdmi_device_t *inno_hdmi)
{
	if (inno_hdmi->chip.hw_fini)
		inno_hdmi->chip.hw_fini(&inno_hdmi->chip);
}

static int innodpu_hdmi_zoomenable_id(int hdmi_id)
{
	int zoom_id = -1;

	switch (hdmi_id) {
		case 0:
			zoom_id = HDMI0_ZOOM_ENABLE;
			break;
		case 1:
			zoom_id = HDMI1_ZOOM_ENABLE;
			break;
		case 2:
			zoom_id = HDMI2_ZOOM_ENABLE;
			break;
		case 3:
			zoom_id = HDMI3_ZOOM_ENABLE;
			break;
		default:
			break;
	}
	return zoom_id;
}

static int inno_hdmi_chip_init(struct hdmi_device_t *inno_hdmi)
{
	chip_type_e plat;
	int retcode = 0;
	int zoom_id = 0;

	zoom_id = innodpu_hdmi_zoomenable_id(inno_hdmi->hdmi_id);

	if ((zoom_id >= 0) && (hal_zoom_is_enable(inno_hdmi->dev->parent, zoom_id) == 1)) {
		inno_hdmi->chip.hal_zoom_enable = false;
		inno_hdmi->zoom_id = zoom_id;
	} else {
		inno_hdmi->chip.hal_zoom_enable = false;
		inno_hdmi->zoom_id = 0;
	}

	inno_hdmi->chip.hal_edid_mode = hal_hdmi_edid_mode(inno_hdmi->dev->parent, inno_hdmi->hdmi_id);
	if (inno_hdmi->chip.hal_edid_mode < 0) {
		innodpu_info(inno_hdmi->dev, DPU_UT_INIT, "hdmi-%d hal_hdmi_edid_mode failed, return %d\n",
			inno_hdmi->hdmi_id, inno_hdmi->chip.hal_edid_mode);
		inno_hdmi->chip.hal_edid_mode = EDID_AUTO_READ;
	}

	if (inno_hdmi->chip.hal_zoom_enable && \
			inno_hdmi->chip.hal_edid_mode == EDID_STR_PUSH) {
			inno_hdmi->chip.hal_edid_mode = EDID_ZOOM_ENABLE;
	}

	innodpu_info(inno_hdmi->dev, DPU_UT_INIT, "hdmi-%d use edid mode:%d\n",
			inno_hdmi->hdmi_id, inno_hdmi->chip.hal_edid_mode);


	plat = hal_get_chiptype(inno_hdmi->dev->parent);
	switch(plat) {
	case CHIP_G1_SOC:
		innodpu_info(inno_hdmi->dev, DPU_UT_INIT, "%s start init g1 soc.\n", inno_hdmi->name);
		inno_hdmi->chip.hal_zoom_enable = false;
		retcode = g1_soc_hdmi_chip_init(&inno_hdmi->chip,
				              inno_hdmi->dev, inno_hdmi->hdmi_id);
		break;
	case CHIP_G0_SOC:
		innodpu_info(inno_hdmi->dev, DPU_UT_INIT, "%s start init g0 soc.\n", inno_hdmi->name);
		retcode = g0_soc_hdmi_chip_init(&inno_hdmi->chip,
				              inno_hdmi->dev, inno_hdmi->hdmi_id);
		break;
	case CHIP_G1P_SOC:
		innodpu_info(inno_hdmi->dev, DPU_UT_INIT, "%s start init g1p soc.\n", inno_hdmi->name);
		inno_hdmi->chip.hal_zoom_enable = false;
		retcode = g1p_soc_hdmi_chip_init(&inno_hdmi->chip,
				              inno_hdmi->dev, inno_hdmi->hdmi_id);
		break;
	case CHIP_G0M_SOC:
		innodpu_info(inno_hdmi->dev, DPU_UT_INIT, "%s start init g0m soc.\n", inno_hdmi->name);
		retcode = g0m_soc_hdmi_chip_init(&inno_hdmi->chip,
				              inno_hdmi->dev, inno_hdmi->hdmi_id);
		break;
	default:
		innodpu_err(inno_hdmi->dev, "%s does not currently support %d platform.\n",
			inno_hdmi->name, plat);
		retcode = -EINVAL;
		break;
	}

	innohdmi_i2c_init(inno_hdmi);

	inno_hdmi->chip.output_mode = innodpu_get_connector_output_mode(inno_hdmi->dev, inno_hdmi->chip.reg_module);
	if (inno_hdmi->chip.output_mode) {
		if (is_output_type_dvi(inno_hdmi->chip.output_mode) ||
			is_output_type_vga(inno_hdmi->chip.output_mode)  ||
			is_output_type_lvds(inno_hdmi->chip.output_mode)) {
			inno_hdmi->chip.max_width = 2048;
			inno_hdmi->chip.max_height = 2048;
		}
		innodpu_info(inno_hdmi->dev, DPU_UT_INIT, "mode_connector:%d convert:%s",
			inno_hdmi->chip.output_mode->mode_connector_item, inno_hdmi->chip.output_mode->convert_name);
	}

	{
		struct hw_board_info board[] = {
			HW_BOARD_INFO_ITEM("ZY", "YF27_1"), //VGA+HDMI2VGA+DP2LVDS
			HW_BOARD_INFO_ITEM("ZY", "YF27_2"), //VGA+HDMI2VGA+DP2VGA
		};
		if (innodpu_odm_pcb_match(inno_hdmi->dev->parent, board, INNO_ARRAY_SIZE(board))) {
			inno_hdmi->chip.max_width = 1920;
			inno_hdmi->chip.max_height = 1080;
		}
	}

	innodpu_info(inno_hdmi->dev, DPU_UT_INIT, "hdmi-%d zoom:%s",
			inno_hdmi->hdmi_id, inno_hdmi->chip.hal_zoom_enable ? "enable" : "disable");

	return retcode;
}

static void inno_hdmi_chip_fini(struct hdmi_device_t *inno_hdmi)
{
	chip_type_e plat;

	plat = hal_get_chiptype(inno_hdmi->dev->parent);
	switch(plat) {
	case CHIP_G1_SOC:
		g1_soc_hdmi_chip_fini(&inno_hdmi->chip);
		break;
	case CHIP_G0_SOC:
		g0_soc_hdmi_chip_fini(&inno_hdmi->chip);
		break;
	case CHIP_G1P_SOC:
		g1p_soc_hdmi_chip_fini(&inno_hdmi->chip);
		break;
	case CHIP_G0M_SOC:
		g0m_soc_hdmi_chip_fini(&inno_hdmi->chip);
		break;
	default:
		innodpu_err(inno_hdmi->dev, "%s does not currently support %d platform.\n",
			inno_hdmi->name, plat);
		break;
	}
	return;
}

/**
 * inno_hdmi_bind - innosilicon hdmi-driver initialization function
 * @dev: hdmi_device_t_info allocated when inno_hdmi_device_t_register
 * parent(dev->parent) is &pci_dev.dev
 * @master: component master
 * @data: point of struct drm_device
 *
 * This function initializes the innosilicon hdmi device
 * 1. alloc and init hdmi handle
 * 2. hdmi hardware init
 * 3. hdmi ddcci supports
 * 4. creation and bonding of HDMI connectors and encoders
 * 5. setup hdmi irq
 *
 * Returns:
 * 0 if it is OK, errno otherwise.
 */

#ifdef CONFIG_DRM_INNO_AUDIO
static void inno_hdmi_audio_enable(struct audio_conn *ac)
{
	struct hdmi_device_t *inno_hdmi = NULL;

	if (!ac || !ac->priv) {
		return ;
	}

	inno_hdmi = (struct hdmi_device_t *)ac->priv;
	if (inno_hdmi && inno_hdmi->chip.hdmi_enable_audio) {
		inno_hdmi->chip.hdmi_enable_audio(&inno_hdmi->chip);
	}
}

static void inno_hdmi_audio_disable(struct audio_conn *ac)
{
	struct hdmi_device_t *inno_hdmi = NULL;

	if (!ac || !ac->priv) {
		return ;
	}

	inno_hdmi = (struct hdmi_device_t *)ac->priv;
	if (inno_hdmi && inno_hdmi->chip.hdmi_disable_audio) {
		inno_hdmi->chip.hdmi_disable_audio(&inno_hdmi->chip);
	}
}
#endif

static int inno_hdmi_bind(struct device *dev,
					  struct device *master, void *data)
{
	int retcode = 0;
	struct drm_device * drm_dev = data;
	struct hdmi_device_t *inno_hdmi = NULL;
	plat_data_t *pdata =  dev_get_platdata(dev);
#ifdef CONFIG_DRM_INNO_AUDIO
	struct audio_conn *pac;
#endif
	struct drm_display_mode *inno_mode = NULL;

	innodpu_info(dev, DPU_UT_INIT,"start\n");
	BUG_ON(!dev);
	BUG_ON(!data);

	inno_hdmi = devm_kmalloc(dev, sizeof(*inno_hdmi), GFP_KERNEL);
	if (!inno_hdmi) {
		innodpu_err(dev, "Alloc hdmi-%d handle failed. Short of memory.\n",
			pdata->dev_idx);
		return -ENOMEM;
	}

	memset(inno_hdmi, 0, sizeof(*inno_hdmi));
	inno_hdmi->hdmi_id = pdata->dev_idx;
	inno_hdmi->dev = inno_get_device(dev);
	inno_hdmi->drm_dev = drm_dev;
	atomic64_set(&inno_hdmi->modesetting, 0);

	inno_hdmi->name = inno_kasprintf(inno_gfp_kernel, "inno-hdmi-%d", inno_hdmi->hdmi_id);
	if (!inno_hdmi->name) {
		innodpu_err(dev, "Alloc hdmi-%d name failed. Short of memory.\n",
			pdata->dev_idx);
		retcode = -ENOMEM;
		goto err_out_name;
	}

	dev_set_drvdata(dev, inno_hdmi);
	innodpu_info(dev, DPU_UT_INIT, "%s start bind.\n", inno_hdmi->name);

	retcode = inno_hdmi_chip_init(inno_hdmi);
	if (retcode) {
		innodpu_err(dev, "%s chip Init failed-%d.\n", inno_hdmi->name, retcode);
		goto err_chip_init;
	}

	retcode = inno_hdmi_hw_init(inno_hdmi);
	if (retcode) {
		innodpu_err(dev, "%s hw Init failed-%d.\n", inno_hdmi->name, retcode);
		goto err_hw_init;
	}

	inno_mode = devm_kmalloc(dev, sizeof(*inno_mode), GFP_KERNEL);
	if (!inno_mode) {
		innodpu_err(dev, "Alloc hdmi-%d mode failed. Short of memory.\n",
			pdata->dev_idx);
		return -ENOMEM;
	}
	memset(inno_mode, 0, sizeof(*inno_mode));
	inno_hdmi->chip.adjusted_mode = inno_mode;

	retcode = inno_hdmi_connector_attach_encoder(inno_hdmi);
	if (retcode) {
		innodpu_err(dev, "%s connector and encoder Init failed-%d.\n",
			inno_hdmi->name, retcode);
		goto err_attach_encoder;
	}

	retcode = hal_set_irq_handler(dev->parent,
		inno_hdmi->chip.hal_module, inno_hdmi_irq_handle, inno_hdmi);
	if (retcode) {
		innodpu_err(dev, "%s irq Init failed-%d.\n", inno_hdmi->name, retcode);
		goto err_enable_irq;
	}

	inno_hdmi->hpdwq = create_singlethread_workqueue("hotplug");
	INIT_DELAYED_WORK(&inno_hdmi->hotplug_work, inno_hdmi_hotplug_work);

	inno_hdmi->scrambwq = create_singlethread_workqueue("scramble");
	INIT_DELAYED_WORK(&inno_hdmi->scramble_work, inno_hdmi_scramble_work);

#ifdef CONFIG_DRM_INNO_AUDIO
	if (hal_getflag_hdmi2dvi(inno_dev_get_parent(dev), inno_hdmi->hdmi_id) <= 0) {
		pac = kzalloc(sizeof(struct audio_conn), GFP_KERNEL);
		if (!pac) {
			innodpu_err(dev, "Alloc audio_conn failed.\n");
			goto err_enable_irq;
		}
		pac->conn_st = 0; //todo
		pac->has_audio = 0; //todo
		pac->dev = (void *)inno_hdmi->dev;
		pac->id = inno_hdmi->hdmi_id;
		pac->type = INNOAUDIO_CONNECTOR_TYPE_HDMI;

		pac->priv = (void *)inno_hdmi;
		pac->enable  = inno_hdmi_audio_enable;
		pac->disable = inno_hdmi_audio_disable;

		if (innoaudio_register_connector(pac)) {
			innodpu_info(dev, DPU_UT_INIT, "register audio_conn failed.\n");
			inno_hdmi->ac = NULL;
			kfree(pac);
		} else {
			innodpu_info(dev, DPU_UT_INIT, "register audio_conn ok.\n");
			inno_hdmi->ac = pac;
		}
	}
#endif

	innodpu_info(dev, DPU_UT_INIT,"end\n");
	return retcode;

err_enable_irq:
	inno_hdmi_cleanup_connector_encoder(inno_hdmi);
err_attach_encoder:
	inno_hdmi_hw_fini(inno_hdmi);
err_hw_init:
	inno_hdmi_chip_fini(inno_hdmi);
err_chip_init:
	if (inno_hdmi->name)
		kfree(inno_hdmi->name);
err_out_name:
	put_device(dev);
	devm_kfree(dev, inno_hdmi);

	return retcode;
}

/**
* inno_hdmi_unbind - innosilicon hdmi-driver initialization function
* @dev: hdmi_device_t_info allocated when inno_hdmi_device_t_register
* parent(dev->parent) is &pci_dev.dev
* @master: component master
* @data: point of struct drm_device
*
* This function deinitializes the innosilicon hdmi device
*
* Once drm_dev_register is called, the connector and encoder
* resources are released by the drm_mode_config_cleanup function!!!
* So we don't need to worry about the release of the connector and encoder here.
*/
static void inno_hdmi_unbind(struct device *dev,
						  struct device *master, void *data)

{
	struct hdmi_device_t *inno_hdmi = NULL;

	innodpu_info(dev, DPU_UT_INIT,"start\n");
	BUG_ON(!dev);
	BUG_ON(!data);

	inno_hdmi = dev_get_drvdata(dev);
	if (!inno_hdmi) {
		innodpu_err(dev, "hdmi handle is NULL\n");
		return;
	}

	hal_set_irq_handler(dev->parent,
		inno_hdmi->chip.hal_module, NULL, NULL);

#ifdef CONFIG_DRM_INNO_AUDIO
	if (inno_hdmi->ac) {
		innoaudio_unregister_connector(inno_hdmi->ac);
		kfree(inno_hdmi->ac);
		inno_hdmi->ac = NULL;
	}
#endif

	cancel_delayed_work_sync(&inno_hdmi->hotplug_work);
	cancel_delayed_work_sync(&inno_hdmi->scramble_work);

	inno_hdmi_hw_fini(inno_hdmi);
	inno_hdmi_chip_fini(inno_hdmi);
	kfree(inno_hdmi->name);

	put_device(dev);
	devm_kfree(dev, inno_hdmi);
	innodpu_info(dev, DPU_UT_INIT,"end\n");
}


static const struct component_ops s_inno_hdmi_ops = {
	.bind = inno_hdmi_bind,
	.unbind = inno_hdmi_unbind,
};

static int inno_hdmi_probe(struct platform_device *pdev)
{
	BUG_ON(!pdev);
	return component_add(&pdev->dev, &s_inno_hdmi_ops);
}

static int inno_hdmi_remove(struct platform_device *pdev)
{
	BUG_ON(!pdev);
	component_del(&pdev->dev, &s_inno_hdmi_ops);
	return 0;
}

static struct platform_device_id s_inno_hdmi_device_id_table[] = {
	{.name = INNO_HDMI_DEVICE_NAME, .driver_data = 0},
	{},
};
MODULE_DEVICE_TABLE(platform, s_inno_hdmi_device_id_table);

static int inno_hdmi_suspend(struct device *dev)
{
	struct hdmi_device_t *inno_hdmi = NULL;

	innodpu_info(dev, DPU_UT_INIT,"start\n");
	BUG_ON(!dev);

	inno_hdmi = dev_get_drvdata(dev);
	if (!inno_hdmi) {
		innodpu_err(dev, "hdmi handle is NULL\n");
		return 0;
	}

	hal_dev_disable_irq(dev->parent, inno_hdmi->chip.hal_module);

	cancel_delayed_work_sync(&inno_hdmi->hotplug_work);
	cancel_delayed_work_sync(&inno_hdmi->scramble_work);

	innodpu_info(dev, DPU_UT_INIT,"end\n");
	return 0;
}

static int inno_hdmi_resume(struct device *dev)
{
	struct hdmi_device_t *inno_hdmi = NULL;

	innodpu_info(dev, DPU_UT_INIT,"start\n");
	BUG_ON(!dev);

	inno_hdmi = dev_get_drvdata(dev);
	if (!inno_hdmi) {
		innodpu_err(dev, "hdmi handle is NULL\n");
		return 0;
	}

	inno_hdmi_hw_init(inno_hdmi);
	drm_helper_hpd_irq_event(inno_hdmi->drm_dev);

	hal_dev_enable_irq(dev->parent, inno_hdmi->chip.hal_module);
	innodpu_info(dev, DPU_UT_INIT,"end\n");
	return 0;
}

static void inno_hdmi_shutdown(struct platform_device *pdev)
{
	struct hdmi_device_t *inno_hdmi = NULL;

	inno_hdmi =  inno_platform_get_drvdata(pdev);
	if (!inno_hdmi) {
		innodpu_err(&pdev->dev, "hdmi handle is NULL\n");
		return;
	}

	inno_hdmi_encoder_mode_disable(&inno_hdmi->encoder);
	inno_hdmi_suspend(&pdev->dev);
}

static const struct dev_pm_ops inno_hdmi_pm_ops = {
	SET_SYSTEM_SLEEP_PM_OPS(inno_hdmi_suspend, inno_hdmi_resume)
};

struct platform_driver g_innogpu_hdmi_driver = {
	.probe = inno_hdmi_probe,
	.remove = inno_hdmi_remove,
	.shutdown = inno_hdmi_shutdown,
	.driver = {
		.name = INNO_HDMI_DEVICE_NAME,
		.pm = &inno_hdmi_pm_ops,
	},
	.id_table = s_inno_hdmi_device_id_table,
};

