/*************************************************************************/ /*!
@File			innovpu_drv.c
@Title			innovpu driver
@Copyright		Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
@Description	innovpu driver
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/device.h>
#include <linux/dma-mapping.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/uaccess.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/version.h>
#include <linux/kthread.h>
#include <linux/pci.h>
#include <linux/kobject.h>
#include <linux/debugfs.h>
#include <linux/vmalloc.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)
#include <linux/sched/signal.h>
#endif

#include "kernel_compatibility.h"
#include "syscommon.h"
#include "osfunc_common.h"
#include "innogpu.h"
#include "inno_misc.h"
#include "innovpu.h"
#include "innovpu_dmabuf.h"
#include "innovpu_common.h"

#define MAX_INTERRUPT_QUEUE (16 * 32)

static u32 vpu_free_buffers(vpu_drv_context *vpu_drv_ctx, void *filp)
{
	vpudrv_buffer_pool_t *pool, *n = NULL;
	vpudrv_buffer_t vb;
	vpu_buf_obj_t obj;
	int found = 0;

	vpu_dbg(vpu_drv_ctx->vpudev, " vpu_free_buffers, filp:%p\n", filp);
	while (1) {
		memset(&vb, 0, sizeof(vpudrv_buffer_t));
		memset(&obj, 0, sizeof(vpudrv_buffer_t));
		spin_lock(vpu_drv_ctx->vpu_lock);
		list_for_each_entry_safe(pool, n, &vpu_drv_ctx->vbp_head, list) {
			if (pool->filp == filp) {
				vb = pool->buf.vb;
				obj = pool->buf.obj;
				found = 1;
				list_del(&pool->list);
				kfree(pool);
				break;
			}
		}
		spin_unlock(vpu_drv_ctx->vpu_lock);

		if (found == 1) {
			found = 0;
			if (vb.fd) {
				atomic64_sub(vb.size, &(vpu_drv_ctx->external_mem));
				vpu_destroy_dmabuf(vpu_drv_ctx, &obj);
			} else if (vb.base) {
				vpu_free_dma_buffer(&vb, vpu_drv_ctx);
			}
	    } else {
			break;
		}
	}
	return 0;
}

static vpu_drv_context * vpu_get_drv_ctx(struct file *filp)
{
	return (vpu_drv_context *)filp->private_data;
}

static int vpu_open(struct inode *inode, struct file *filp)
{
	vpu_drv_context *vpu_drv_ctx;

	vpu_drv_ctx = container_of((struct cdev *)inode->i_cdev, vpu_drv_context, vpucdev);

	spin_lock(vpu_drv_ctx->vpu_lock);
	vpu_drv_ctx->vpu_context.open_count++;
	spin_unlock(vpu_drv_ctx->vpu_lock);

	filp->private_data = vpu_drv_ctx;

	return 0;
}

/*static u32 vpu_ioctl(struct inode *inode, struct file *filp, u_int cmd, u_long arg) // for kernel 2.6.9*/
static long vpu_ioctl(struct file *filp, u_int cmd, u_long arg)
{
	int ret = 0;

	vpu_drv_context *vpu_drv_ctx = vpu_get_drv_ctx(filp);
	if (!vpu_drv_ctx) {
		vpu_prerr("%s vpu_drv_ctx is NULL\n", __func__);
		return -EINVAL;
	}

	switch (cmd) {
		case VDI_IOCTL_DMA_MAP_HOSTMEM:
			{
				vpudrv_buffer_t vb;
				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_DMA_MAP_HOSTMEM\n");
				ret = inno_copy_from_user(&vb, (const vpudrv_buffer_t* __user)arg, sizeof(vpudrv_buffer_t));
				if (ret) {
					return -EFAULT;
				}
				if ((ret = down_interruptible(vpu_drv_ctx->vpu_sem)) == 0) {
					ret = vpu_dma_map_hostmem(vpu_drv_ctx, &vb);
					up(vpu_drv_ctx->vpu_sem);
					if (ret) {
						break;
					}
				}
				ret = inno_copy_to_user((void* __user)arg, &vb,sizeof(vpudrv_buffer_t));
				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_DMA_MAP_HOSTMEM\n");
			}
			break;
		case VDI_IOCTL_DMA_UNMAP_HOSTMEM:
			{
				vpudrv_buffer_t vb;
				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_DMA_UNMAP_HOSTMEM\n");
				ret = inno_copy_from_user(&vb, (const vpudrv_buffer_t* __user)arg, sizeof(vpudrv_buffer_t));
				if (ret) {
					return -EFAULT;
				}
				if ((ret = down_interruptible(vpu_drv_ctx->vpu_sem)) == 0) {
					vpu_dma_unmap_hostmem(vpu_drv_ctx, &vb);
					up(vpu_drv_ctx->vpu_sem);
				}
				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_DMA_UNMAP_HOSTMEM\n");
			}
			break;
		case VDI_IOCTL_ALLOCATE_PHYSICAL_MEMORY:
			{
				vpudrv_buffer_pool_t *vbp;

				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_ALLOCATE_PHYSICAL_MEMORY\n");

				if ((ret = down_interruptible(vpu_drv_ctx->vpu_sem)) == 0) {
					vbp = kzalloc(sizeof(*vbp), GFP_KERNEL);
					if (!vbp) {
						up(vpu_drv_ctx->vpu_sem);
						return -ENOMEM;
					}

					ret = inno_copy_from_user(&(vbp->buf.vb), (vpudrv_buffer_t *)arg, sizeof(vpudrv_buffer_t));
					if (ret) {
						kfree(vbp);
						up(vpu_drv_ctx->vpu_sem);
						return -EFAULT;
					}

					ret = vpu_alloc_dma_buffer(&(vbp->buf.vb), vpu_drv_ctx);
					if (ret != 0) {
						ret = -ENOMEM;
						kfree(vbp);
						up(vpu_drv_ctx->vpu_sem);
						break;
					}

					ret = copy_to_user((void __user *)arg, &(vbp->buf.vb), sizeof(vpudrv_buffer_t));
					if (ret) {
						vpu_free_dma_buffer(&(vbp->buf.vb), vpu_drv_ctx);
						kfree(vbp);
						ret = -EFAULT;
						up(vpu_drv_ctx->vpu_sem);
						break;
					}

					vbp->filp = filp;
					spin_lock(vpu_drv_ctx->vpu_lock);
					list_add(&vbp->list, &vpu_drv_ctx->vbp_head);
					spin_unlock(vpu_drv_ctx->vpu_lock);

					up(vpu_drv_ctx->vpu_sem);
				}
				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_ALLOCATE_PHYSICAL_MEMORY\n");
			}
			break;
		case VDI_IOCTL_FREE_PHYSICALMEMORY:
			{
				vpudrv_buffer_pool_t *vbp = NULL, *n = NULL;
				vpudrv_buffer_t vb;
				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_FREE_PHYSICALMEMORY\n");

				if ((ret = down_interruptible(vpu_drv_ctx->vpu_sem)) == 0) {
					ret = inno_copy_from_user(&vb, (vpudrv_buffer_t *)arg, sizeof(vpudrv_buffer_t));
					if (ret) {
						up(vpu_drv_ctx->vpu_sem);
						return -EACCES;
					}

					if (vb.base) {
						vpu_free_dma_buffer(&vb, vpu_drv_ctx);
					}

					spin_lock(vpu_drv_ctx->vpu_lock);
					list_for_each_entry_safe(vbp, n, &vpu_drv_ctx->vbp_head, list)
					{
						if (vbp->buf.vb.base == vb.base) {
							list_del(&vbp->list);
							kfree(vbp);
							break;
						}
					}
					spin_unlock(vpu_drv_ctx->vpu_lock);

					up(vpu_drv_ctx->vpu_sem);
				}
				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_FREE_PHYSICALMEMORY\n");

			}
			break;
		case VDI_IOCTL_GET_RESERVED_VIDEO_MEMORY_INFO:
			break;
		case VDI_IOCTL_DMABUF_EXPORT:
			{
				vpudrv_buffer_pool_t *vbp;
				int fd;

				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_DMABUF_EXPORT\n");
				vbp = kzalloc(sizeof(*vbp), GFP_KERNEL);
				if (!vbp) {
					return -ENOMEM;
				}

				ret = inno_copy_from_user(&(vbp->buf.vb), (vpudrv_buffer_t *)arg, sizeof(vpudrv_buffer_t));
				if (ret) {
					kfree(vbp);
					return -EFAULT;
				}

				if ((ret = down_interruptible(vpu_drv_ctx->vpu_sem)) == 0) {
					fd = vpu_export_dmabuf(vpu_drv_ctx, &(vbp->buf.obj));
					if (fd < 0) {
						ret = -ENOMEM;
						kfree(vbp);
						up(vpu_drv_ctx->vpu_sem);
						break;
					}
					up(vpu_drv_ctx->vpu_sem);
				} else {
					kfree(vbp);
					return -EFAULT;
				}

				ret = copy_to_user((void __user *)arg, &(vbp->buf.vb), sizeof(vpudrv_buffer_t));
				if (ret) {
					vpu_free_dma_buffer(&(vbp->buf.obj.vb), vpu_drv_ctx);
					vpu_destroy_dmabuf(vpu_drv_ctx, &(vbp->buf.obj));
					kfree(vbp);
					ret = -EFAULT;
					break;
				}

				vbp->filp = filp;
				atomic64_add(vbp->buf.vb.size, &(vpu_drv_ctx->external_mem));
				spin_lock(vpu_drv_ctx->vpu_lock);
				list_add(&vbp->list, &vpu_drv_ctx->vbp_head);
				spin_unlock(vpu_drv_ctx->vpu_lock);

				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_DMABUF_EXPORT\n");
			}
			break;
		case VDI_IOCTL_DMABUF_IMPORT:
			{
				vpudrv_buffer_pool_t *vbp;
				vpudrv_buffer_t *vb;

				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_DMABUF_IMPORT\n");
				vbp = kzalloc(sizeof(*vbp), GFP_KERNEL);
				if (!vbp) {
					return -ENOMEM;
				}

				ret = inno_copy_from_user(&(vbp->buf.vb), (vpudrv_buffer_t *)arg, sizeof(vpudrv_buffer_t));
				if (ret) {
					kfree(vbp);
					return -EFAULT;
				}
				vb = vpu_import_dmabuf(vpu_drv_ctx, &(vbp->buf.obj));
				if (IS_ERR_OR_NULL(vb)) {
					ret = PTR_ERR(vb);
					kfree(vbp);
					break;
				}

				ret = copy_to_user((void __user *)arg, vb, sizeof(vpudrv_buffer_t));
				if (ret) {
					kfree(vbp);
					ret = -EFAULT;
					vpu_destroy_dmabuf(vpu_drv_ctx, &(vbp->buf.obj));
					break;
				}

				vbp->filp = filp;
				atomic64_add(vbp->buf.vb.size, &(vpu_drv_ctx->external_mem));
				spin_lock(vpu_drv_ctx->vpu_lock);
				list_add(&vbp->list, &vpu_drv_ctx->vbp_head);
				spin_unlock(vpu_drv_ctx->vpu_lock);

				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_DMABUF_IMPORT\n");
			}
			break;
		case VDI_IOCTL_DMABUF_DESTROY:
			{
				vpudrv_buffer_pool_t *vbp = NULL, *n = NULL;
				vpudrv_buffer_t vb;
				vpu_buf_obj_t obj;
				int found = 0;

				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_DMABUF_DESTROY\n");
				ret = inno_copy_from_user(&vb, (vpudrv_buffer_t *)arg, sizeof(vpudrv_buffer_t));
				if (ret < 0) {
					return -EFAULT;
				}

				spin_lock(vpu_drv_ctx->vpu_lock);
				list_for_each_entry_safe(vbp, n, &vpu_drv_ctx->vbp_head, list) {
					if ((vbp->buf.vb.fd == vb.fd) && (vbp->filp == filp) && (vbp->buf.vb.phys_addr == vb.phys_addr)) {
						found = 1;
						obj = vbp->buf.obj;
						list_del(&vbp->list);
						kfree(vbp);
						break;
					}
				}
				spin_unlock(vpu_drv_ctx->vpu_lock);

				if (found == 1) {
					atomic64_sub(obj.vb.size, &(vpu_drv_ctx->external_mem));
					vpu_destroy_dmabuf(vpu_drv_ctx, &obj);
				}
				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_DMABUF_DESTROY\n");
			}
			break;
		case VDI_IOCTL_DMAFENCE_SIGNAL:
			{
#ifdef DMA_FENCE_SYNC_SUPPORT
				sync_event_t ev;
				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_DMAFENCE_SIGNAL\n");

				ret = inno_copy_from_user(&ev, (sync_event_t *)arg, sizeof(ev));
				if (ret < 0) {
					return -EFAULT;
				}
				ret = vpu_dma_fence_signal(ev.fd, ev.event);
				if (ret) {
					vpu_error(vpu_drv_ctx->vpudev, "vpu hw reset failed\n");
					return -EFAULT;
				}
				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_DMAFENCE_SIGNAL\n");
#endif
			}
			break;
		case VDI_IOCTL_DMAFENCE_CREATE:
			{

#ifdef DMA_FENCE_SYNC_SUPPORT
				sync_event_t ev;
				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_DMAFENCE_CREATE\n");

				ret = inno_copy_from_user(&ev, (sync_event_t *)arg, sizeof(ev));
				if (ret < 0) {
					return -EFAULT;
				}
				ret = vpu_dma_fence_create(ev.fd, ev.event);
				if (ret != 0) {
					vpu_error(vpu_drv_ctx->vpudev, "vpu_dma_fence_create failed\n");
					return -EFAULT;
				}
				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_DMAFENCE_CREATE\n");
#endif
			}
			break;
		case VDI_IOCTL_WAIT_INTERRUPT:
			{
				vpudrv_intr_info_t info;
#ifdef SUPPORT_MULTI_INST_INTR
				u32 intr_inst_index;
				u32 intr_reason_in_q;
				u32 interrupt_flag_in_q;
#endif
				struct vpu_context_t *vpu_ctx = &vpu_drv_ctx->vpu_context;

				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_WAIT_INTERRUPT\n");

				ret = inno_copy_from_user(&info, (vpudrv_intr_info_t *)arg, sizeof(vpudrv_intr_info_t));
				if (ret != 0)
				{
					return -EFAULT;
				}
#ifdef SUPPORT_MULTI_INST_INTR
				intr_inst_index = info.intr_inst_index;

				intr_reason_in_q = 0;
				interrupt_flag_in_q = kfifo_out_spinlocked(vpu_drv_ctx->interrupt_pending_q[intr_inst_index],
					&intr_reason_in_q, sizeof(u32), vpu_drv_ctx->kfifo_lock);
				if (interrupt_flag_in_q > 0)
				{
					vpu_ctx->interrupt_reason[intr_inst_index] = intr_reason_in_q;
					vpu_dbg(vpu_drv_ctx->vpudev, " Interrupt Remain : intr_inst_index=%d, intr_reason_in_q=0x%x, interrupt_flag_in_q=%d\n",
						intr_inst_index, intr_reason_in_q, interrupt_flag_in_q);
					goto INTERRUPT_REMAIN_IN_QUEUE;
				}

				ret = wait_event_interruptible_timeout(*(vpu_drv_ctx->interrupt_wait_q[intr_inst_index]),
					vpu_drv_ctx->interrupt_flag[intr_inst_index] != 0, msecs_to_jiffies(info.timeout));

#endif
				if (!ret) {
					//vpu_error(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_WAIT_INTERRUPT timeout = %d \n", info.timeout);
					ret = -ETIME;
					break;
				}

				if (signal_pending(current)) {
					ret = -ERESTARTSYS;
					break;
				}

#ifdef SUPPORT_MULTI_INST_INTR
				intr_reason_in_q = 0;
				interrupt_flag_in_q = kfifo_out_spinlocked(vpu_drv_ctx->interrupt_pending_q[intr_inst_index],
					&intr_reason_in_q, sizeof(u32), vpu_drv_ctx->kfifo_lock);
				if (interrupt_flag_in_q > 0) {
					vpu_ctx->interrupt_reason[intr_inst_index] = intr_reason_in_q;
				}
				else {
					vpu_ctx->interrupt_reason[intr_inst_index] = 0;
				}
				vpu_dbg(vpu_drv_ctx->vpudev, " inst_index(%d), interrupt_flag(%d), reason(0x%08llx)\n", intr_inst_index,
					vpu_drv_ctx->interrupt_flag[intr_inst_index], vpu_ctx->interrupt_reason[intr_inst_index]);

INTERRUPT_REMAIN_IN_QUEUE:
				info.intr_reason = vpu_ctx->interrupt_reason[intr_inst_index];
				vpu_drv_ctx->interrupt_flag[intr_inst_index] = 0;
				vpu_ctx->interrupt_reason[intr_inst_index] = 0;
#endif
				ret = copy_to_user((void __user *)arg, &info, sizeof(vpudrv_intr_info_t));
				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_WAIT_INTERRUPT\n");
				if (ret != 0)
				{
					return -EFAULT;
				}
			}
			break;
		case VDI_IOCTL_SET_CLOCK_GATE:
			{
				u32 clkgate;

				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_SET_CLOCK_GATE\n");
				//if (get_user(clkgate, (u32 __user *) arg))
				if (inno_copy_from_user(&clkgate,  (void *)arg, sizeof(u32)))
					return -EFAULT;
				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_SET_CLOCK_GATE\n");

			}
			break;
		case VDI_IOCTL_GET_INSTANCE_POOL:
			{
				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_GET_INSTANCE_POOL\n");
				if ((ret = down_interruptible(vpu_drv_ctx->vpu_sem)) == 0) {

					if (vpu_drv_ctx->instance_pool.base != 0) {

						ret = copy_to_user((void __user *)arg, &vpu_drv_ctx->instance_pool, sizeof(vpudrv_buffer_t));
						if (ret != 0)
							ret = -EFAULT;
					} else {

						ret = inno_copy_from_user(&vpu_drv_ctx->instance_pool, (vpudrv_buffer_t *)arg, sizeof(vpudrv_buffer_t));
						if (ret == 0) {
							vpu_drv_ctx->instance_pool.base = (u64)vmalloc(PAGE_ALIGN(vpu_drv_ctx->instance_pool.size));
							vpu_drv_ctx->instance_pool.phys_addr = vpu_drv_ctx->instance_pool.base;
							if (vpu_drv_ctx->instance_pool.base != 0) {
								/*clearing memory*/
								memset((void *)vpu_drv_ctx->instance_pool.base, 0x0, PAGE_ALIGN(vpu_drv_ctx->instance_pool.size));
								ret = copy_to_user((void __user *)arg, &vpu_drv_ctx->instance_pool, sizeof(vpudrv_buffer_t));
								if (ret == 0) {
									/* success to get memory for instance pool */
									up(vpu_drv_ctx->vpu_sem);
									break;
								}
							}
						}

						ret = -EFAULT;
					}
					up(vpu_drv_ctx->vpu_sem);
				}

				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_GET_INSTANCE_POOL\n");
			}
			break;
		case VDI_IOCTL_GET_COMMON_MEMORY:
			{
				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_GET_COMMON_MEMORY\n");
				if (vpu_drv_ctx->common_memory.base != 0) {
					ret = copy_to_user((void __user *)arg, &vpu_drv_ctx->common_memory, sizeof(vpudrv_buffer_t));
					if (ret != 0) {
						ret = -EFAULT;
						break;
					}
				} else {
					ret = inno_copy_from_user(&vpu_drv_ctx->common_memory, (vpudrv_buffer_t *)arg, sizeof(vpudrv_buffer_t));
					if (ret != 0) {
						ret = -EFAULT;
						break;
					}
					ret = vpu_alloc_dma_buffer(&vpu_drv_ctx->common_memory, vpu_drv_ctx);
					if (ret != 0) {
						ret = -EFAULT;
						break;
					}
 					ret = copy_to_user((void __user *)arg, &vpu_drv_ctx->common_memory, sizeof(vpudrv_buffer_t));
					if (ret != 0) {
						vpu_free_dma_buffer(&vpu_drv_ctx->common_memory, vpu_drv_ctx);
						ret = -EFAULT;
						break;
					}
				}
				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_GET_COMMON_MEMORY\n");
			}
			break;
		case VDI_IOCTL_GET_PVRIC_MEMORY:
			{
				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_GET_PVRIC_MEMORY\n");
				if (vpu_drv_ctx->pvric_memory.base != 0) {
					ret = copy_to_user((void __user *)arg, &vpu_drv_ctx->pvric_memory, sizeof(vpudrv_buffer_t));
					if (ret != 0) {
						ret = -EFAULT;
						break;
					}
				} else {
					ret = inno_copy_from_user(&vpu_drv_ctx->pvric_memory, (vpudrv_buffer_t *)arg, sizeof(vpudrv_buffer_t));
					if (ret != 0) {
						ret = -EFAULT;
						break;
					}
					ret = vpu_alloc_dma_buffer(&vpu_drv_ctx->pvric_memory, vpu_drv_ctx);
					if (ret != 0) {
						ret = -EFAULT;
						break;
					}
					ret = copy_to_user((void __user *)arg, &vpu_drv_ctx->pvric_memory, sizeof(vpudrv_buffer_t));
					if (ret != 0) {
						vpu_free_dma_buffer(&vpu_drv_ctx->pvric_memory, vpu_drv_ctx);
						ret = -EFAULT;
						break;
					}
				}
				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_GET_PVRIC_MEMORY\n");
			}
			break;
		case VDI_IOCTL_OPEN_INSTANCE:
			{
				vpudrv_inst_info_t inst_info;
				vpudrv_instanace_list_t *vil, *n = NULL;
				int codecIndex = -1;

				vil = kzalloc(sizeof(*vil), GFP_KERNEL);
				if (!vil) {
					return -ENOMEM;
				}
				if (inno_copy_from_user(&inst_info, (vpudrv_inst_info_t *)arg, sizeof(vpudrv_inst_info_t))) {
					kfree(vil);
					return -EFAULT;
				}

				codecIndex = vpu_capture_inst_index(vpu_drv_ctx, inst_info.core_idx);
				if (codecIndex < 0) {
					kfree(vil);
					return -EFAULT;
				}
				inst_info.inst_idx = codecIndex;

				vil->inst_idx = inst_info.inst_idx;
				vil->core_idx = inst_info.core_idx;
				vil->filp = (void*)filp;

				spin_lock(vpu_drv_ctx->vpu_lock);
				list_add(&vil->list, &vpu_drv_ctx->inst_list_head);

				inst_info.inst_open_count = 0; /* counting the current open instance number */
				list_for_each_entry_safe(vil, n, &vpu_drv_ctx->inst_list_head, list)
				{
					if (vil->core_idx == inst_info.core_idx) {
						inst_info.inst_open_count++;
					}
				}
				vpu_drv_ctx->vpu_open_ref_count += 1; /* flag just for that vpu is in opened or closed */
				spin_unlock(vpu_drv_ctx->vpu_lock);

				ret = vpu_mutex_lock(inst_info.core_idx, inst_info.inst_idx, VPUDRV_MUTEX_VPU, vpu_drv_ctx);
				if (ret != 0 && ret != 1) {
					return -EFAULT;
				}
#ifdef SUPPORT_MULTI_INST_INTR
				kfifo_reset(vpu_drv_ctx->interrupt_pending_q[inst_info.inst_idx]);
#endif
				vpu_mutex_unlock(inst_info.core_idx, inst_info.inst_idx, VPUDRV_MUTEX_VPU, vpu_drv_ctx);

				if (copy_to_user((void __user *)arg, &inst_info, sizeof(vpudrv_inst_info_t))) {
					return -EFAULT;
				}

				inno_spin_lock(vpu_drv_ctx->inst_lock_info_lock[inst_info.inst_idx]);
				INIT_LIST_HEAD(&vpu_drv_ctx->inst_lock_info_head[inst_info.inst_idx]);
				inno_spin_unlock(vpu_drv_ctx->inst_lock_info_lock[inst_info.inst_idx]);

				vpu_dbg(vpu_drv_ctx->vpudev, " VDI_IOCTL_OPEN_INSTANCE core_idx=%d, inst_idx=%d, vpu_open_ref_count=%d, inst_open_count=%d\n",
					(u32)inst_info.core_idx, (u32)inst_info.inst_idx, vpu_drv_ctx->vpu_open_ref_count, inst_info.inst_open_count);
			}
			break;
		case VDI_IOCTL_CLOSE_INSTANCE:
			{
				vpudrv_inst_info_t inst_info;
				vpudrv_instanace_list_t *vil = NULL, *n = NULL;
				vpu_lock_info_list_t *vlil = NULL, *vliln = NULL;
				uint64_t vpuLoad = 0;
				u32 found = 0;

				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_CLOSE_INSTANCE\n");
				if (inno_copy_from_user(&inst_info, (vpudrv_inst_info_t *)arg, sizeof(vpudrv_inst_info_t)))
					return -EFAULT;

				spin_lock(vpu_drv_ctx->vpu_lock);
				list_for_each_entry_safe(vil, n, &vpu_drv_ctx->inst_list_head, list)
				{
					if (vil->inst_idx == inst_info.inst_idx && vil->core_idx == inst_info.core_idx) {
						vpuLoad = vil->inst_param.width * vil->inst_param.height * vil->inst_param.frameRate;
						atomic64_sub(vpuLoad, &(vpu_drv_ctx->statistic_load));
						list_del(&vil->list);
						kfree(vil);
						found = 1;
						break;
					}
				}

				if (0 == found) {
					spin_unlock(vpu_drv_ctx->vpu_lock);
					return -EINVAL;
				}

				inst_info.inst_open_count = 0; /* counting the current open instance number */
				vpu_drv_ctx->vpu_open_ref_count -= 1; /* flag just for that vpu is in opened or closed */
				list_for_each_entry_safe(vil, n, &vpu_drv_ctx->inst_list_head, list)
				{
					if (vil->core_idx == inst_info.core_idx) {
						inst_info.inst_open_count++;
					}
				}
				spin_unlock(vpu_drv_ctx->vpu_lock);

				inno_spin_lock(vpu_drv_ctx->inst_lock_info_lock[inst_info.inst_idx]);
				list_for_each_entry_safe(vlil, vliln, &vpu_drv_ctx->inst_lock_info_head[inst_info.inst_idx], list) {
					list_del(&vlil->list);
					kfree(vlil);
				}
				inno_spin_unlock(vpu_drv_ctx->inst_lock_info_lock[inst_info.inst_idx]);

				ret = vpu_mutex_lock(inst_info.core_idx, inst_info.inst_idx, VPUDRV_MUTEX_VPU, vpu_drv_ctx);
				if (ret != 0 && ret != 1) {
					return -EFAULT;
				}
#ifdef SUPPORT_MULTI_INST_INTR
				kfifo_reset(vpu_drv_ctx->interrupt_pending_q[inst_info.inst_idx]);
#endif
				vpu_mutex_unlock(inst_info.core_idx, inst_info.inst_idx, VPUDRV_MUTEX_VPU, vpu_drv_ctx);

				if (vpu_release_inst_index(vpu_drv_ctx, inst_info.core_idx, inst_info.inst_idx) < 0) {
					return -EFAULT;
				}

				if (copy_to_user((void __user *)arg, &inst_info, sizeof(vpudrv_inst_info_t)))
					return -EFAULT;

				vpu_dbg(vpu_drv_ctx->vpudev, " VDI_IOCTL_CLOSE_INSTANCE core_idx=%d, inst_idx=%d, vpu_open_ref_count=%d, inst_open_count=%d\n",
					(u32)inst_info.core_idx, (u32)inst_info.inst_idx, vpu_drv_ctx->vpu_open_ref_count, inst_info.inst_open_count);
			}
			break;
		case VDI_IOCTL_GET_INSTANCE_NUM:
			{
				vpudrv_inst_info_t inst_info;
				vpudrv_instanace_list_t *vil = NULL, *n = NULL;
				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_GET_INSTANCE_NUM\n");

				ret = inno_copy_from_user(&inst_info, (vpudrv_inst_info_t *)arg, sizeof(vpudrv_inst_info_t));
				if (ret != 0)
					break;

				spin_lock(vpu_drv_ctx->vpu_lock);
				inst_info.inst_open_count = 0;
				list_for_each_entry_safe(vil, n, &vpu_drv_ctx->inst_list_head, list)
				{
					if (vil->core_idx == inst_info.core_idx) {
						inst_info.inst_open_count++;
					}
				}
				spin_unlock(vpu_drv_ctx->vpu_lock);

				ret = copy_to_user((void __user *)arg, &inst_info, sizeof(vpudrv_inst_info_t));
				vpu_dbg(vpu_drv_ctx->vpudev, " VDI_IOCTL_GET_INSTANCE_NUM core_idx=%d, inst_idx=%d, open_count=%d\n",
					(u32)inst_info.core_idx, (u32)inst_info.inst_idx, inst_info.inst_open_count);
			}
			break;
		case VDI_IOCTL_RESET:
			{
				ret = vpu_hw_reset(vpu_drv_ctx);
				if (ret) {
					vpu_error(vpu_drv_ctx->vpudev, "vpu hw reset failed\n");
					return -EINVAL;
				}
			}
			break;
		case VDI_IOCTL_DMA_TRANSFER:
			{
				vpu_dma_info_t info;
				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_DMA_TRANSFER \n");

				ret = inno_copy_from_user(&info, (vpu_dma_info_t *)arg, sizeof(vpu_dma_info_t));
				if (ret) {
					return -EFAULT;
				}
                vpu_dbg(vpu_drv_ctx->vpudev, " direction %d src domain %d dst domain %d\n",
                            info.direction, info.src_vb.domain, info.dest_vb.domain);
				ret = vpu_dma_transfer(vpu_drv_ctx, info);
				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_DMA_TRANSFER\n");
			}
			break;
		case VDI_IOCTL_DMA_TRANSFER_FAST:
			{
				vpu_dma_info_t info;
				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_DMA_TRANSFER_FAST \n");
				ret = copy_from_user(&info, (vpu_dma_info_t *)arg, sizeof(vpu_dma_info_t));
				if (ret) {
					return -EFAULT;
				}
				vpu_dbg(vpu_drv_ctx->vpudev, " direction %d src domain %d dst domain %d\n",
							info.direction, info.src_vb.domain, info.dest_vb.domain);
				ret = vpu_dma_transfer_fast(vpu_drv_ctx, info);
				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_DMA_TRANSFER_FAST\n");
			}
			break;
		case VDI_IOCTL_GET_REGISTER_INFO:
			{
				vpu_dbg(vpu_drv_ctx->vpudev, "[+]VDI_IOCTL_GET_REGISTER_INFO\n");
				ret = copy_to_user((void __user *)arg, &vpu_drv_ctx->vpu_register, sizeof(vpudrv_buffer_t));
				if (ret != 0) {
					ret = -EFAULT;
				}
				vpu_dbg(vpu_drv_ctx->vpudev, "[-]VDI_IOCTL_GET_REGISTER_INFO vpu_register.phys_addr==0x%llx, vpu_register.virt_addr=0x%llx, vpu_register.size=%d\n",
					vpu_drv_ctx->vpu_register.phys_addr , vpu_drv_ctx->vpu_register.virt_addr, vpu_drv_ctx->vpu_register.size);
			}
			break;
		case VDI_IOCTL_GET_FREE_MEM_SIZE:
			{
				u64 size;

				size = vpu_get_free_mem_size(vpu_drv_ctx);
				if (copy_to_user((void __user *)arg, &size, sizeof(u64))) {
					return -EFAULT;
				}
			}
			break;
		case VDI_IOCTL_GET_CHIP_INFO:
			{
				if (copy_to_user((void __user *)arg, &vpu_drv_ctx->chip_info, sizeof(vpudrv_chip_info_t))) {
					return -EFAULT;
				}
			}
			break;
		case VDI_IOCTL_LOCK_ENQUEUE:
			{
				vpu_lock_info_t lock_info;

				if (inno_copy_from_user(&lock_info, (vpu_lock_info_t *)arg, sizeof(vpu_lock_info_t))) {
					return -EFAULT;
				}

				ret = vpu_lock_enqueue(vpu_drv_ctx, &lock_info);
			}
			break;
		case VDI_IOCTL_LOCK_DEQUEUE:
			{
				u64 instIndex;
				if (inno_copy_from_user(&instIndex, (u64 *)arg, sizeof(u64))) {
					return -EFAULT;
				}

				ret = vpu_lock_dequeue(vpu_drv_ctx, instIndex);
			}
			break;
		case VDI_IOCTL_TRY_LOCK:
			{
				u64 instIndex;
				if (inno_copy_from_user(&instIndex, (u64 *)arg, sizeof(u64))) {
					return -EFAULT;
				}

				ret = vpu_try_lock(vpu_drv_ctx, instIndex);
			}
			break;
		case VDI_IOCTL_UNLOCK:
			{
				u64 instIndex;
				if (inno_copy_from_user(&instIndex, (u64 *)arg, sizeof(u64))) {
					return -EFAULT;
				}

				ret = vpu_unlock(vpu_drv_ctx, instIndex);
			}
			break;
		case VDI_IOCTL_CREATE_INSTANCE:
			{
				u64 instIndex;
				if (inno_copy_from_user(&instIndex, (u64 *)arg, sizeof(u64))) {
					return -EFAULT;
				}

				ret = vpu_create_instance(vpu_drv_ctx, instIndex);
			}
			break;
		case VDI_IOCTL_DESTROY_INSTANCE:
			{
				u64 instIndex;
				if (inno_copy_from_user(&instIndex, (u64 *)arg, sizeof(u64))) {
					return -EFAULT;
				}

				ret = vpu_destroy_instance(vpu_drv_ctx, instIndex);
			}
			break;
		case VDI_IOCTL_INCREASE_LOAD:
			{
				vpudrv_instanace_list_t *vil = NULL, *n = NULL;
				vpu_instance_param_t instParam;
				uint64_t vpuLoad = 0;
				if (inno_copy_from_user(&instParam, (vpu_instance_param_t *)arg, sizeof(vpu_instance_param_t))) {
					return -EFAULT;
				}

				inno_spin_lock(vpu_drv_ctx->vpu_lock);
				list_for_each_entry_safe(vil, n, &vpu_drv_ctx->inst_list_head, list)
				{
					if (vil->filp == filp) {
						vil->inst_param.width = instParam.width;
						vil->inst_param.height = instParam.height;
						vil->inst_param.frameRate = instParam.frameRate;
						vpuLoad = instParam.width * instParam.height * instParam.frameRate;
						atomic64_add(vpuLoad, &(vpu_drv_ctx->statistic_load));
						break;
					}
				}
				inno_spin_unlock(vpu_drv_ctx->vpu_lock);
			}
			break;
		default:
			{
				vpu_error(vpu_drv_ctx->vpudev, " No such IOCTL, cmd is %d\n", cmd);
			}
			break;
	}

	return ret;
}

static ssize_t vpu_read(struct file *filp, char __user *buf, size_t len, loff_t *ppos)
{
	return -1;
}

static ssize_t vpu_write(struct file *filp, const char __user *buf, size_t len, loff_t *ppos)
{

	vpu_drv_context *vpu_drv_ctx = vpu_get_drv_ctx(filp);

	/* vpu_dbg(vpu_drv_ctx->vpudev, " vpu_write len=%d\n", (u32)len); */
	if (!buf) {
		vpu_error(vpu_drv_ctx->vpudev, " vpu_write buf = NULL error \n");
		return -EFAULT;
	}

	if (len == sizeof(vpu_bit_firmware_info_t))	{
		vpu_bit_firmware_info_t *bit_firmware_info;

		bit_firmware_info = kmalloc(sizeof(vpu_bit_firmware_info_t), GFP_KERNEL);
		if (!bit_firmware_info) {
			vpu_error(vpu_drv_ctx->vpudev, " vpu_write  bit_firmware_info allocation error \n");
			return -EFAULT;
		}

		if (inno_copy_from_user(bit_firmware_info, buf, len)) {
			vpu_error(vpu_drv_ctx->vpudev, " vpu_write inno_copy_from_user error for bit_firmware_info\n");
			kfree(bit_firmware_info);
			return -EFAULT;
		}

		if (bit_firmware_info->size == sizeof(vpu_bit_firmware_info_t)) {
			vpu_dbg(vpu_drv_ctx->vpudev, " vpu_write set bit_firmware_info coreIdx=0x%x, reg_base_offset=0x%x size=0x%x, bit_code[0]=0x%x\n",
					bit_firmware_info->core_idx, (u32)bit_firmware_info->reg_base_offset, bit_firmware_info->size,
					bit_firmware_info->bit_code[0]);

			if (bit_firmware_info->core_idx > MAX_NUM_VPU_CORE) {
				vpu_error(vpu_drv_ctx->vpudev, " vpu_write coreIdx[%d] is exceeded than MAX_NUM_VPU_CORE[%d]\n",
					bit_firmware_info->core_idx, MAX_NUM_VPU_CORE);
				return -ENODEV;
			}
			memcpy(&(vpu_drv_ctx->bit_firmware_info[bit_firmware_info->core_idx]), bit_firmware_info, sizeof(vpu_bit_firmware_info_t));
			kfree(bit_firmware_info);
			return len;
		}
		kfree(bit_firmware_info);
	}

	return -1;
}

static int vpu_release(struct inode *inode, struct file *filp)
{
#ifdef SUPPORT_MULTI_INST_INTR
	u32 i;
#endif
	vpu_drv_context *vpu_drv_ctx = vpu_get_drv_ctx(filp);
	vpu_dbg(vpu_drv_ctx->vpudev, " vpu_release\n");

	vpu_free_lock(vpu_drv_ctx, filp);
	/* found and free the not closed instance by user applications */
	vpu_free_instances(vpu_drv_ctx, filp);
	/* found and free the not handled buffer by user applications */
	vpu_free_buffers(vpu_drv_ctx, filp);

	spin_lock(vpu_drv_ctx->vpu_lock);
	vpu_drv_ctx->vpu_context.open_count--;
	if (vpu_drv_ctx->vpu_context.open_count == 0) {
#ifdef SUPPORT_MULTI_INST_INTR
		for (i = 0; i < MAX_NUM_INSTANCE; i++) {
			kfifo_reset(vpu_drv_ctx->interrupt_pending_q[i]);
		}
#endif
		/*
		   Don't release instance_pool.base here to simply pm notifier.
		   This buffer will be released @ innovpu_remove.
		 */
	}
	spin_unlock(vpu_drv_ctx->vpu_lock);

	return 0;
}

static int vpu_fasync(int fd, struct file *filp, int mode)
{
	vpu_drv_context *vpu_drv_ctx = vpu_get_drv_ctx(filp);
	struct vpu_context_t *dev = &(vpu_drv_ctx->vpu_context);
	return fasync_helper(fd, filp, mode, &dev->async_queue);
}

static int vpu_map_to_register(struct file *fp, struct vm_area_struct *vm)
{
	u64 pfn;
	vpu_drv_context *vpu_drv_ctx = vpu_get_drv_ctx(fp);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 3, 0))
	vm->vm_flags |= VM_IO | VM_RESERVED;
#else
	vm_flags_set(vm, vm->vm_flags | VM_IO | VM_RESERVED);
#endif
	vm->vm_page_prot = pgprot_noncached(vm->vm_page_prot);
	pfn =  vpu_drv_ctx->vpu_register.phys_addr >> PAGE_SHIFT;

	return remap_pfn_range(vm, vm->vm_start, pfn, vm->vm_end-vm->vm_start, vm->vm_page_prot) ? -EAGAIN : 0;
}

static int vpu_map_to_physical_memory(struct file *fp, struct vm_area_struct *vm)
{
	vpudrv_buffer_pool_t *vbp = NULL, *n = NULL;
	u64 pfn = vm->vm_pgoff;
	vpu_drv_context *vpu_drv_ctx = vpu_get_drv_ctx(fp);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 3, 0))
	vm->vm_flags |= VM_IO | VM_RESERVED;
#else
	vm_flags_set(vm, vm->vm_flags | VM_IO | VM_RESERVED);
#endif
	vm->vm_page_prot = pgprot_noncached(vm->vm_page_prot);

  	//TODO Update after having a good optimization plan
	if ((vm->vm_pgoff & VPU_FOUR_BYTES_ANDOPERATION) == ((vpu_drv_ctx->common_memory.phys_addr >> PAGE_SHIFT) & VPU_FOUR_BYTES_ANDOPERATION))
	{
		pfn = vpu_drv_ctx->common_memory.phys_addr >> PAGE_SHIFT;
	} else {
		spin_lock(vpu_drv_ctx->vpu_lock);
		list_for_each_entry_safe(vbp, n, &vpu_drv_ctx->vbp_head, list)
		{
			if ((vm->vm_pgoff & VPU_FOUR_BYTES_ANDOPERATION) == ((vbp->buf.vb.phys_addr >> PAGE_SHIFT) & VPU_FOUR_BYTES_ANDOPERATION))
			{
				pfn = vbp->buf.vb.phys_addr >> PAGE_SHIFT;
				break;
			}
		}
		spin_unlock(vpu_drv_ctx->vpu_lock);
	}
	return remap_pfn_range(vm, vm->vm_start, pfn, vm->vm_end-vm->vm_start, vm->vm_page_prot) ? -EAGAIN : 0;
}

static int vpu_map_to_instance_pool_memory(struct file *fp, struct vm_area_struct *vm)
{
	int ret;
	long long length = vm->vm_end - vm->vm_start;
	u64 start = vm->vm_start;
	vpu_drv_context *vpu_drv_ctx = vpu_get_drv_ctx(fp);
	char *vmalloc_area_ptr = (char *)vpu_drv_ctx->instance_pool.base;
	u64 pfn;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 3, 0))
	vm->vm_flags |= VM_RESERVED;
#else
	vm_flags_set(vm, vm->vm_flags | VM_RESERVED);
#endif

	/* loop over all pages, map it page individually */
	while (length > 0)
	{
		pfn = vmalloc_to_pfn(vmalloc_area_ptr);
		if ((ret = remap_pfn_range(vm, start, pfn, PAGE_SIZE, PAGE_SHARED)) < 0) {
			return ret;
		}
		start += PAGE_SIZE;
		vmalloc_area_ptr += PAGE_SIZE;
		length -= PAGE_SIZE;
	}

	return 0;
}

/*!
 * @brief memory map interface for vpu file operation
 * @return  0 on success or negative error code on error
 */
static int vpu_mmap(struct file *fp, struct vm_area_struct *vm)
{
	vpu_drv_context *vpu_drv_ctx = vpu_get_drv_ctx(fp);

	if (vm->vm_pgoff == 0)
		return vpu_map_to_instance_pool_memory(fp, vm);

  	//TODO Update after having a good optimization plan
	if ((vm->vm_pgoff & VPU_FOUR_BYTES_ANDOPERATION) == ((vpu_drv_ctx->vpu_register.phys_addr >> PAGE_SHIFT) & VPU_FOUR_BYTES_ANDOPERATION))
		return vpu_map_to_register(fp, vm);

	return vpu_map_to_physical_memory(fp, vm);
}

static struct file_operations vpu_fops = {
	.owner = THIS_MODULE,
	.open = vpu_open,
	.read = vpu_read,
	.write = vpu_write,
	/*.ioctl = vpu_ioctl, // for kernel 2.6.9 of*/
	.unlocked_ioctl = vpu_ioctl,
	.compat_ioctl = vpu_ioctl,
	.release = vpu_release,
	.fasync = vpu_fasync,
	.mmap = vpu_mmap,
};

static void size_to_string(u64 size, char* str, int str_len)
{
    if (size > 1024*1024) {
		snprintf(str, str_len, "%llu MB", size >> 20);
    } else if (size > 1024) {
        snprintf(str, str_len, "%llu KB", size >> 10);
    } else {
		snprintf(str, str_len, "%llu B", size);
    }
}

static ssize_t vpu_read_file(struct file *file, char __user *user_buf,
				     size_t count, loff_t *ppos)
{
	ssize_t ret = 0;
	u64 visible_mem = 0;
	u64 invisible_mem = 0;
	u64 external_mem = 0;
	u64 vpuLoad = 0;
	char *buf;
	char tmp_buf[32] = {0};
	vpu_drv_context *vpu_drv_ctx = (vpu_drv_context *)file->private_data;
	struct vram_stats visible_stats = {0};
	struct vram_stats invisible_stats = {0};
	struct role_target role = {0};

	if (!vpu_drv_ctx)
		return -EINVAL;

	visible_mem = atomic64_read(&(vpu_drv_ctx->visible_mem));
	invisible_mem = atomic64_read(&(vpu_drv_ctx->invisible_mem));
	external_mem = atomic64_read(&(vpu_drv_ctx->external_mem));
	vpuLoad = atomic64_read(&(vpu_drv_ctx->statistic_load));

	role.vram_role = HAL_VRAM_ROLE_VPU;
	role.id = vpu_drv_ctx->vpu_id;
	role.sub_id = 0;
	if (hal_get_vram_stats(vpu_drv_ctx->parent, &role, true, &visible_stats) != 0) {
		vpu_error(vpu_drv_ctx->dev, "hal_get_vram_stats failed\n");
		return -EINVAL;
	}

	if (vpu_drv_ctx->chip_info.bar_size & INNO_GPU_BAR_SIZE_SMALL) {
		if (hal_get_vram_stats(vpu_drv_ctx->parent, &role, false, &invisible_stats) != 0) {
			vpu_error(vpu_drv_ctx->dev, "hal_get_vram_stats failed\n");
			return -EINVAL;
		}
	}

	buf = (char *)__get_free_page(GFP_KERNEL);
	if (!buf)
		return -ENOMEM;

	ret += sprintf(buf, "RunInstanceNum: %d\n", vpu_drv_ctx->vpu_open_ref_count);
	ret += sprintf(buf + ret, "VpuLoadUsage: %llu%%\n", vpuLoad * 100 / VPU_MAX_LOAD_SIZE);
	size_to_string(visible_mem, tmp_buf, sizeof(tmp_buf));
	ret += sprintf(buf + ret, "VisibleMem: %s\n", tmp_buf);
	size_to_string(visible_stats.total_size, tmp_buf, sizeof(tmp_buf));
	ret += sprintf(buf + ret, "VisibleTotalMem: %s\n", tmp_buf);
	size_to_string(visible_stats.free_size, tmp_buf, sizeof(tmp_buf));
	ret += sprintf(buf + ret, "VisibleTotalFreeMem: %s\n", tmp_buf);
	if (visible_stats.total_size > 0) {
		ret += sprintf(buf + ret, "VisibleTotalUsage: %llu%%\n", (visible_stats.total_size - visible_stats.free_size) * 100 / visible_stats.total_size);
	}

	size_to_string(invisible_mem, tmp_buf, sizeof(tmp_buf));
	ret += sprintf(buf + ret, "InvisibleMem: %s\n", tmp_buf);
	size_to_string(invisible_stats.total_size, tmp_buf, sizeof(tmp_buf));
	ret += sprintf(buf + ret, "InvisibleTotalMem: %s\n", tmp_buf);
	size_to_string(invisible_stats.free_size, tmp_buf, sizeof(tmp_buf));
	ret += sprintf(buf + ret, "InvisibleTotalFreeMem: %s\n", tmp_buf);
	if (invisible_stats.total_size > 0) {
		ret += sprintf(buf + ret, "InvisibleTotalUsage: %llu%%\n", (invisible_stats.total_size - invisible_stats.free_size) * 100 / invisible_stats.total_size);
	}

	size_to_string(external_mem, tmp_buf, sizeof(tmp_buf));
	ret += sprintf(buf + ret, "ImportMem: %s\n", tmp_buf);
	ret = simple_read_from_buffer(user_buf, count, ppos, buf, strlen(buf));

	free_page((unsigned long)buf);
	return ret;
}

static int vpu_irq[] = {HAL_INTERRUPT_CODEC_0,
                        HAL_INTERRUPT_CODEC_1,
                        HAL_INTERRUPT_CODEC_2,
                        HAL_INTERRUPT_CODEC_3,
                        HAL_INTERRUPT_CODEC_4,
                        HAL_INTERRUPT_CODEC_5};

static const struct file_operations debugfs_fops = {
	.open = simple_open,
	.read = vpu_read_file,
};

static int vpu_create_debugfs(vpu_drv_context *vpu_drv_ctx)
{
	/*create vpu debug sys*/
	vpu_drv_ctx->vpu_dir = debugfs_create_dir(vpu_drv_ctx->vpuname, NULL);
	if (IS_ERR(vpu_drv_ctx->vpu_dir)) {
		vpu_error(vpu_drv_ctx->dev, "failed to create vpu dir debugfs\n");
		goto ERR_CREATE_DIR;
	}

	vpu_drv_ctx->vpu_node = debugfs_create_file("vpu_status", 0444,
							vpu_drv_ctx->vpu_dir,
							vpu_drv_ctx, &debugfs_fops);
	if (IS_ERR(vpu_drv_ctx->vpu_node)) {
		vpu_error(vpu_drv_ctx->dev, "failed to create vpu node debugfs\n");
		goto ERR_DEBUGFS_FILE;
	}

	return 0;

ERR_DEBUGFS_FILE:
	debugfs_remove(vpu_drv_ctx->vpu_dir);

ERR_CREATE_DIR:
	return -1;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 2, 0)
static char *vpu_devnode(const struct device *dev, umode_t *mode)
#else
static char *vpu_devnode(struct device *dev, umode_t *mode)
#endif
{
	if (mode) {
		*mode = 0666;
	}

	return NULL;
}

static int vpu_create_device(vpu_drv_context *vpu_drv_ctx)
{
	int ret = -1;
	struct device *vpu_device;

	if (!vpu_drv_ctx) {
		vpu_prerr("%s: vpu_drv_ctx is NULL \n", __func__);
		return -1;
	}

	//alloc cdev
	vpu_drv_ctx->cdevid = (dev_t *)kzalloc(sizeof(dev_t), GFP_KERNEL);
	ret = alloc_chrdev_region(vpu_drv_ctx->cdevid, 0, 1, vpu_drv_ctx->vpuname);
	if (ret < 0) {
		vpu_error(vpu_drv_ctx->dev, "alloc_chrdev_region failed(ret=%d)\n",ret);
		goto ERR_ALLOC_CDEV;
	}

	//init cdev
	vpu_drv_ctx->vpucdev.owner = THIS_MODULE;
	cdev_init(&vpu_drv_ctx->vpucdev, &vpu_fops);

	//register cdev
	ret = cdev_add(&vpu_drv_ctx->vpucdev, *vpu_drv_ctx->cdevid, 1);
	if (ret < 0) {
		vpu_error(vpu_drv_ctx->dev, "cdev_add failed (ret=%d)\n",ret);
		goto ERR_ADD_CDEV;
	}

	//create class
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 4, 0)
	vpu_drv_ctx->vpucls = class_create(vpu_drv_ctx->vpuname);
#else
	vpu_drv_ctx->vpucls = class_create(THIS_MODULE, vpu_drv_ctx->vpuname);
#endif
	if (IS_ERR(vpu_drv_ctx->vpucls)) {
		vpu_error(vpu_drv_ctx->dev, "class_create failed\n");
		goto ERR_CLASS_CREATE;
	}

	//set /dev/vpu* user mode 0666
	vpu_drv_ctx->vpucls->devnode = vpu_devnode;

	//create vpu device
	vpu_device = device_create(vpu_drv_ctx->vpucls, NULL, *vpu_drv_ctx->cdevid, NULL, vpu_drv_ctx->vpuname);
	if (IS_ERR(vpu_device)) {
		vpu_error(vpu_drv_ctx->dev, "device_create failed\n");
		goto ERR_DEVICE_CREATE;
	}
	vpu_drv_ctx->vpudev = vpu_device;

	return 0;

ERR_DEVICE_CREATE:
	class_destroy(vpu_drv_ctx->vpucls);

ERR_CLASS_CREATE:
	cdev_del(&vpu_drv_ctx->vpucdev);

ERR_ADD_CDEV:
	unregister_chrdev_region(*vpu_drv_ctx->cdevid, 1);

ERR_ALLOC_CDEV:
	kfree(vpu_drv_ctx->cdevid);

	return -ENOMEM;
}

static void vpu_destroy_device(vpu_drv_context *vpu_drv_ctx)
{
	if (!vpu_drv_ctx) {
		vpu_prerr("%s: vpu_drv_ctx is NULL \n", __func__);
		return;
	}

	device_destroy(vpu_drv_ctx->vpucls, *vpu_drv_ctx->cdevid);
	class_destroy(vpu_drv_ctx->vpucls);
	cdev_del(&vpu_drv_ctx->vpucdev);
	unregister_chrdev_region(*vpu_drv_ctx->cdevid, 1);
	kfree(vpu_drv_ctx->cdevid);
}

static void innovpu_handle_irq(void *data)
{
	vpu_drv_context *vpu_drv_ctx = (vpu_drv_context *)data;

	if (!vpu_drv_ctx) {
		vpu_prerr("%s: vpu_drv_ctx is NULL \n", __func__);
		return;
	}

	vpu_handle_irq(vpu_drv_ctx);
}

static int vpu_unset_irq(vpu_drv_context *vpu_drv_ctx)
{
	int ret = 0;

	if (!vpu_drv_ctx) {
		vpu_prerr("%s: vpu_drv_ctx is NULL \n", __func__);
		return -1;
	}

	ret = hal_dev_disable_irq(vpu_drv_ctx->parent, vpu_drv_ctx->irq_handler_num);
	if (ret) {
		vpu_error(vpu_drv_ctx->dev, "hal_dev_disable_irq failed (ret=%d)\n", ret);
	}

	return ret;
}

static int vpu_set_irq(vpu_drv_context *vpu_drv_ctx)
{
	int ret = 0;

	if (!vpu_drv_ctx) {
		vpu_prerr("%s: vpu_drv_ctx is NULL \n", __func__);
		return -1;
	}

	ret = hal_set_irq_handler(vpu_drv_ctx->parent, vpu_drv_ctx->irq_handler_num, innovpu_handle_irq, vpu_drv_ctx);
	if (ret) {
		vpu_error(vpu_drv_ctx->dev, "failed to set codec interrupt handler (ret=%d)\n", ret);
		return -1;
	}

	ret = hal_dev_enable_irq(vpu_drv_ctx->parent, vpu_drv_ctx->irq_handler_num);
	if (ret) {
		vpu_error(vpu_drv_ctx->dev, "failed to enable codec interrupts (ret=%d)\n", ret);
		return -1;
	}

	return 0;
}

static int vpu_get_hwinfo(vpu_drv_context *vpu_drv_ctx)
{
	if (!vpu_drv_ctx) {
		vpu_prerr("%s: vpu_drv_ctx is NULL \n", __func__);
		return -1;
	}

	vpu_drv_ctx->chip_info.chip_type = hal_get_chiptype(vpu_drv_ctx->parent);
	if (0 == vpu_drv_ctx->chip_info.chip_type) {
		vpu_error(vpu_drv_ctx->dev, "hal_get_chiptype unknown\n");
		return -1;
	}

	vpu_drv_ctx->chip_info.vpu_num = hal_get_dev_nums(vpu_drv_ctx->parent, DEV_VPU);
	vpu_drv_ctx->chip_info.vf_id = hal_get_vfid(vpu_drv_ctx->parent);
	if (hal_has_inv_mem(vpu_drv_ctx->parent)) {
		vpu_drv_ctx->chip_info.bar_size = INNO_GPU_BAR_SIZE_SMALL;
	} else {
		vpu_drv_ctx->chip_info.bar_size = INNO_GPU_BAR_SIZE_4G;
	}

	vpu_drv_ctx->irq_handler_num = vpu_irq[vpu_drv_ctx->vpu_id];
	vpu_drv_ctx->chip_info.host_bar2_base_addr = hal_get_ddrbase(vpu_drv_ctx->parent);
	vpu_drv_ctx->chip_info.host_bar2_end_addr = vpu_drv_ctx->chip_info.host_bar2_base_addr + hal_get_ddr_bar_len(vpu_drv_ctx->parent);
	vpu_drv_ctx->chip_info.vram_dev_base_addr = hal_get_vram_dev_base(vpu_drv_ctx->parent);
	vpu_drv_ctx->chip_info.vram_dev_end_addr = vpu_drv_ctx->chip_info.vram_dev_base_addr + hal_get_ddr_bar_len(vpu_drv_ctx->parent);
	vpu_drv_ctx->is_sharing_gpu_heap = hal_is_sharing_gpu_heap(vpu_drv_ctx->parent);
	/* 由于硬复位之后才能正常读取vpu寄存器值，但是硬复位需要product code会判断，所以根据芯片type手动赋值 */
	vpu_drv_ctx->product_code = vpu_get_product_code_manual(vpu_drv_ctx);

	vpu_info(vpu_drv_ctx->dev, "vpu%d product code=0x%x\n", vpu_drv_ctx->vpu_id, vpu_drv_ctx->product_code);

	return 0;
}

static void vpu_deinit_context(vpu_drv_context *vpu_drv_ctx)
{
	int i;

	if (!vpu_drv_ctx) {
		vpu_prerr("%s: vpu_drv_ctx is NULL \n", __func__);
		return;
	}

	kfree(vpu_drv_ctx->vpu_sem);
	inno_spinlock_free(vpu_drv_ctx->vpu_lock);
	inno_spinlock_free(vpu_drv_ctx->kfifo_lock);
	inno_spinlock_free(vpu_drv_ctx->inst_index_lock);

	for (i = 0; i < MAX_NUM_INSTANCE; i++) {
		kfifo_free(vpu_drv_ctx->interrupt_pending_q[i]);
		kfree(vpu_drv_ctx->interrupt_pending_q[i]);
		kfree(vpu_drv_ctx->interrupt_wait_q[i]);
		inno_spinlock_free(vpu_drv_ctx->inst_lock_info_lock[i]);
		inno_waitqueue_head_free(vpu_drv_ctx->inst_lock_wait_q[i]);
	}
}

static void vpu_init_context(vpu_drv_context *ctx)
{
	int i;
	int ret;

	if (!ctx) {
		vpu_prerr("%s: vpu_drv_ctx is NULL \n", __func__);
		return;
	}

	atomic64_set(&(ctx->invisible_mem), 0);
	atomic64_set(&(ctx->visible_mem), 0);
	atomic64_set(&(ctx->external_mem), 0);
	atomic64_set(&(ctx->statistic_load), 0);

	ctx->common_memory.base = 0;
	ctx->instance_pool.base = 0;
	ctx->pvric_memory.base = 0;

	ctx->kfifo_lock = inno_spinlock_alloc();
	ctx->vpu_lock = inno_spinlock_alloc();
	ctx->inst_index_lock = inno_spinlock_alloc();

	ctx->vpu_sem = (struct semaphore *)kzalloc(sizeof(struct semaphore), GFP_KERNEL);
	sema_init(ctx->vpu_sem, 1);

	INIT_LIST_HEAD(&ctx->vbp_head);
	INIT_LIST_HEAD(&ctx->pm_buf_head);
	INIT_LIST_HEAD(&ctx->inst_list_head);

	for (i = 0; i < MAX_NUM_INSTANCE; i++) {
		ctx->interrupt_wait_q[i] = (wait_queue_head_t *)kzalloc(sizeof(wait_queue_head_t), GFP_KERNEL);
		init_waitqueue_head(ctx->interrupt_wait_q[i]);

		ctx->interrupt_pending_q[i] = (kfifo_t *)kzalloc(sizeof(kfifo_t), GFP_KERNEL);
		ret = kfifo_alloc(ctx->interrupt_pending_q[i], MAX_INTERRUPT_QUEUE * sizeof(u32), GFP_KERNEL);
		if (ret) {
			vpu_error(ctx->dev, " kfifo_alloc failed 0x%x\n", ret);
		}

		INIT_LIST_HEAD(&ctx->inst_lock_info_head[i]);
		ctx->inst_lock_info_lock[i] = inno_spinlock_alloc();
		ctx->inst_lock_wait_q[i] = inno_waitqueue_head_alloc();
		init_waitqueue_head(ctx->inst_lock_wait_q[i]);
		ctx->inst_lock_wake_up_flag[i] = 0;
	}
}

#ifdef CONFIG_PM_SLEEP
int vpu_notifier(struct notifier_block *nb, unsigned long action, void *data)
{
	vpu_drv_context *ctx = container_of(nb, vpu_drv_context, pm_notifier);
	int ret = 0;

	switch (action) {
	case PM_HIBERNATION_PREPARE:
	case PM_SUSPEND_PREPARE:
		if (ctx->instance_pool.base != 0) {
			ret = vpu_mutex_lock(0, (u32)MAX_NUM_INSTANCE, VPUDRV_MUTEX_VPU, ctx);
			if (ret != 0) {
				vpu_error(ctx->dev, "%s: vpu_mutex_lock failed preparing...\n", __func__);
			}
			vpu_info(ctx->dev, "%s: vpu_mutex_lock.\n", __func__);
		}
		break;

	case PM_POST_HIBERNATION:
	case PM_POST_SUSPEND:
		if (ctx->instance_pool.base != 0) {
			vpu_mutex_unlock(0, (u32)MAX_NUM_INSTANCE, VPUDRV_MUTEX_VPU, ctx);
			vpu_info(ctx->dev, "%s: vpu_mutex_unlock.\n", __func__);
		}
		break;

	case PM_POST_RESTORE:
	case PM_RESTORE_PREPARE:
	default:
		vpu_warn(ctx->dev, "%s: untested action %d...\n", __func__, action);
		break;
	}

	return NOTIFY_DONE;
}
#endif

static int innovpu_probe(struct platform_device *pdev)
{
	int ret = 0;
	u32 vpu_id, pdev_id;
	plat_data_t *pdata;
	struct resource *res = NULL;
	vpu_drv_context *vpu_drv_ctx;

	if (!pdev) {
		vpu_prerr("%s platform_device is null", __func__);
		return -EINVAL;
	}

	pdata = dev_get_platdata(&pdev->dev);
	if (!pdata) {
		vpu_prerr("%s dev_get_platdata failed", __func__);
		return -EINVAL;
	}
	vpu_id = pdata->dev_idx;
	pdev_id = pdata->pdev_idx;

	vpu_info(&(pdev->dev), "vpu_id: %d:%d, irq: %d\n", pdev_id, vpu_id, vpu_irq[vpu_id]);

	vpu_drv_ctx = kzalloc(sizeof(vpu_drv_context), GFP_KERNEL);
	if (unlikely(!vpu_drv_ctx)) {
		vpu_error(&(pdev->dev), "no memory for vpu driver context!\n");
		return -ENOMEM;
	}

	vpu_drv_ctx->dev = &(pdev->dev);
	vpu_drv_ctx->parent = pdev->dev.parent;
	vpu_drv_ctx->vpu_id = vpu_id;
	vpu_drv_ctx->hostmem_buffers = inno_idr_kalloc( );

	res = platform_get_resource_byname(pdev, IORESOURCE_MEM, "vpu-regs");
	if (res) {/* if platform driver is implemented */
		vpu_drv_ctx->vpu_register.phys_addr = res->start;
		vpu_drv_ctx->vpu_register.virt_addr = (u64)ioremap_nocache(res->start, res->end - res->start);
		vpu_drv_ctx->vpu_register.size = (res->end - res->start + 1);
		vpu_info(&(pdev->dev), "vpu-regs paddr=0x%llx, vaddr=0x%llx\n",
			vpu_drv_ctx->vpu_register.phys_addr ,  vpu_drv_ctx->vpu_register.virt_addr);
	} else {
		vpu_info(&(pdev->dev), "vpu-regs paddr=0x%llx, vaddr=0x%llx\n",
			vpu_drv_ctx->vpu_register.phys_addr ,  vpu_drv_ctx->vpu_register.virt_addr);
		ret = -EINVAL;
		goto ERR_GET_RESOURCE;
	}
	vpu_drv_ctx->vpu_register.domain = INNO_VPU_GEM_DOMAIN_CPU;

	vpu_init_context(vpu_drv_ctx);

	ret = vpu_get_hwinfo(vpu_drv_ctx);
	if (ret) {
		vpu_error(&(pdev->dev), "vpu_get_hwinfo failed\n");
		goto ERR_GET_HWINFO;
	}

	/* move vpu res init from xx_soc.c, xx_soc.c only vpu platform register */
	ret = vpu_hw_res_init(vpu_drv_ctx);
	if (ret) {
		vpu_error(&(pdev->dev), "vpu_hw_res_init failed\n");
		goto ERR_GET_HWINFO;
	}

	/* read product code from reg after hw reset */
	if (vpu_drv_ctx->product_code != vpu_get_product_code(vpu_drv_ctx)) {
		vpu_error(&(pdev->dev), "vpu get product code failed from reg\n");
		goto ERR_GET_HWINFO;
	}

	ret = vpu_set_irq(vpu_drv_ctx);
	if (ret) {
		vpu_error(&(pdev->dev), "vpu_set_irq failed\n");
		goto ERR_GET_HWINFO;
	}

#ifdef USE_CODEC_NAME//TODO...
	vpu_set_vpudev_name(vpu_drv_ctx->dev, vpu_drv_ctx->product_code, (char *)vpu_drv_ctx->vpuname);
#else
	sprintf((char *)vpu_drv_ctx->vpuname, "vpu%d", (pdev_id * vpu_drv_ctx->chip_info.vpu_num + vpu_id));
#endif

	ret = vpu_create_device(vpu_drv_ctx);
	if (ret) {
		vpu_error(&(pdev->dev), "failed to create vpu device\n");
		goto ERR_CREATE_DEVICE;
	}

	ret = vpu_create_debugfs(vpu_drv_ctx);
	if (ret) {
		vpu_error(&(pdev->dev), "failed to create vpu debugfs\n");
		goto ERR_CREATE_DEBUGFS;
	}

	platform_set_drvdata(pdev, vpu_drv_ctx);

#ifdef CONFIG_PM_SLEEP
	vpu_drv_ctx->pm_notifier.notifier_call = vpu_notifier;
	register_pm_notifier(&vpu_drv_ctx->pm_notifier);
#endif

	vpu_dbg(&(pdev->dev), " %s probe success\n", vpu_drv_ctx->vpuname);
	return 0;

ERR_CREATE_DEBUGFS:
	vpu_destroy_device(vpu_drv_ctx);

ERR_CREATE_DEVICE:
#ifdef USE_CODEC_NAME
	vpu_reset_vpudev_num(vpu_drv_ctx->dev, vpu_drv_ctx->product_code);
#endif

	vpu_unset_irq(vpu_drv_ctx);

ERR_GET_HWINFO:
	vpu_deinit_context(vpu_drv_ctx);

	if (vpu_drv_ctx->vpu_register.virt_addr) {
		iounmap((void *) vpu_drv_ctx->vpu_register.virt_addr);
	}

ERR_GET_RESOURCE:
	kfree(vpu_drv_ctx);
	vpu_error(&(pdev->dev), "vpu driver probe failed\n");

	return ret;
}

#ifdef VPU_SUPPORT_PLATFORM_DRIVER_REGISTER
static int innovpu_remove(struct platform_device *pdev)
{
	vpu_drv_context *vpu_drv_ctx = platform_get_drvdata(pdev);
	if (!vpu_drv_ctx) {
		vpu_prerr("%s: vpu_drv_ctx is NULL \n", __func__);
		return -EINVAL;
	}

	vpu_dbg(vpu_drv_ctx->vpudev, "vpu_remove\n");

	if (vpu_drv_ctx->vpu_node)
		debugfs_remove(vpu_drv_ctx->vpu_node);

	if (vpu_drv_ctx->vpu_dir)
		debugfs_remove(vpu_drv_ctx->vpu_dir);

	if (vpu_drv_ctx->instance_pool.base) {
		vfree((const void *)vpu_drv_ctx->instance_pool.base);
		vpu_drv_ctx->instance_pool.base = 0;
	}

	if (vpu_drv_ctx->common_memory.base) {
		vpu_free_dma_buffer(&vpu_drv_ctx->common_memory, vpu_drv_ctx);
		vpu_drv_ctx->common_memory.base = 0;
	}

	if (vpu_drv_ctx->pvric_memory.base) {
		vpu_free_dma_buffer(&vpu_drv_ctx->pvric_memory, vpu_drv_ctx);
		vpu_drv_ctx->pvric_memory.base = 0;
	}

	if (vpu_drv_ctx->hostmem_buffers) {
		/* free all host mem buffer*/
		inno_idr_foreach(vpu_drv_ctx->hostmem_buffers, vpu_idr_dma_unmap_hostmem, vpu_drv_ctx->parent);
		/* release all internal memory from an IDR */
		inno_idr_destroy(vpu_drv_ctx->hostmem_buffers);
		/* release IDR */
		inno_idr_free(vpu_drv_ctx->hostmem_buffers);
		vpu_drv_ctx->hostmem_buffers = NULL;
	}
	vpu_destroy_device(vpu_drv_ctx);

#ifdef USE_CODEC_NAME
	vpu_reset_vpudev_num(vpu_drv_ctx->dev, vpu_drv_ctx->product_code);
#endif

	vpu_unset_irq(vpu_drv_ctx);

	vpu_deinit_context(vpu_drv_ctx);

	if (vpu_drv_ctx->vpu_register.virt_addr)
		iounmap((void *)vpu_drv_ctx->vpu_register.virt_addr);

#ifdef CONFIG_PM_SLEEP
	unregister_pm_notifier(&vpu_drv_ctx->pm_notifier);
#endif

	kfree(vpu_drv_ctx);

	return 0;
}

static void innovpu_shutdown(struct platform_device *pdev)
{
	vpu_drv_context *vpu_drv_ctx = platform_get_drvdata(pdev);

	vpu_prinfo("%s:%d\n", __func__, __LINE__);

	vpu_poweroff(vpu_drv_ctx);
}

static struct platform_device_id s_inno_vpu_platform_device_id_table[] = {
	{ .name = INNO_CODEC_DEVICE_NAME, .driver_data = 0 },
	{ },
};
MODULE_DEVICE_TABLE(platform, s_inno_vpu_platform_device_id_table);

static int innovpu_suspend(struct device *dev)
{
	struct platform_device *pdev;
	vpu_drv_context *vpu_drv_ctx;

	vpu_prinfo("%s:%d\n", __func__, __LINE__);

	pdev = container_of(dev, struct platform_device, dev);
	vpu_drv_ctx = platform_get_drvdata(pdev);

	return vpu_suspend(vpu_drv_ctx);
}

static int innovpu_resume(struct device *dev)
{
	struct platform_device *pdev;
	vpu_drv_context *vpu_drv_ctx;

	vpu_prinfo("%s:%d\n", __func__, __LINE__);

	pdev = container_of(dev, struct platform_device, dev);
	vpu_drv_ctx = platform_get_drvdata(pdev);

	return vpu_resume(vpu_drv_ctx);
}

static int innovpu_freeze(struct device *dev)
{
	vpu_prinfo("%s:%d\n", __func__, __LINE__);

	return innovpu_suspend(dev);
}

static int innovpu_thaw(struct device *dev)
{
	struct platform_device *pdev = container_of(dev, struct platform_device, dev);
	vpu_drv_context *vpu_drv_ctx = platform_get_drvdata(pdev);

	vpu_prinfo("%s:%d\n", __func__, __LINE__);

	return vpu_thaw(vpu_drv_ctx);
}

static int innovpu_poweroff(struct device *dev)
{
	struct platform_device *pdev = container_of(dev, struct platform_device, dev);
	vpu_drv_context *vpu_drv_ctx = platform_get_drvdata(pdev);

	vpu_prinfo("%s:%d\n", __func__, __LINE__);

	return vpu_poweroff(vpu_drv_ctx);
}

static int innovpu_restore(struct device *dev)
{
	vpu_prinfo("%s:%d\n", __func__, __LINE__);

	return innovpu_resume(dev);
}

static const struct dev_pm_ops innovpu_pm_ops = {
#ifdef CONFIG_PM
	.suspend = innovpu_suspend,
	.resume = innovpu_resume,
	.freeze = innovpu_freeze,
	.restore = innovpu_restore,
#endif
	.thaw = innovpu_thaw,
	.poweroff = innovpu_poweroff,
};

static struct platform_driver innovpu_driver = {
	.driver = {
		.name = INNO_CODEC_DEVICE_NAME,
		.pm = &innovpu_pm_ops,
	},
	.probe = innovpu_probe,
	.remove = innovpu_remove,
	.shutdown = innovpu_shutdown,
	.id_table = s_inno_vpu_platform_device_id_table,
};
#endif /* VPU_SUPPORT_PLATFORM_DRIVER_REGISTER */

#ifdef CONFIG_DRM_INNO_VPU
int innovpu_init(void)
#else
static int __init vpu_init(void)
#endif
{
	int ret;
	gpu_extern_vram_flag(2);
	ret = platform_driver_register(&innovpu_driver);
	return ret;
}

#ifdef CONFIG_DRM_INNO_VPU
void innovpu_exit(void)
#else
static void __exit vpu_exit(void)
#endif
{
	platform_driver_unregister(&innovpu_driver);
	return;
}

#ifndef CONFIG_DRM_INNO_VPU
module_init(vpu_init);
module_exit(vpu_exit);

MODULE_AUTHOR("Innosilicon Technologies Ltd. <support@innosilicon.com.cn>");
MODULE_DESCRIPTION("VPU linux driver");
MODULE_LICENSE("Dual BSD/GPL");
#endif

