#include <linux/cdev.h>
#include <linux/uaccess.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include "innodma.h"
#include "gtest_dma_kernel.h"
#include "inno_lock.h"
#include "inno_mm.h"
#include "kernel_compatibility.h"
#include "hal_interface.h"
#include "inno_debug.h"
#include "inno_plat_dev.h"

MODULE_LICENSE("GPL");
#define GTEST_DMA_TEST	"gtest_dma_test"
#define GTEST_DMA_TEST_CLASS	"gtest_dma_test_class"
#define GTEST_DMA_TEST_DEV	"gtest_dma_test_dev"
#define GTEST_DMA_CDEV_SIZE	512

#define DMA_IOC_MAGIC	't'
#define DMA_DDR2SYS		_IO(DMA_IOC_MAGIC, 1)
#define DMA_SYS2DDR		_IO(DMA_IOC_MAGIC, 2)
#define DMA_DDR2DDR		_IO(DMA_IOC_MAGIC, 3)

static struct cdev gtest_dma_cdev;
static dev_t gtest_dma_cdev_id = 0;
static int gtest_dma_major = 0;
static int gtest_dma_minor = 0;
static struct class *gtest_dma_test_class = NULL;
static struct device *gtest_dma_test_device = NULL;
static char* dma_buf = NULL;
bool is_inv_mem = false;
u64 sys_dev_addr = COMMON_VRAM_ADDR_START;

pci_dma_dev_info_t g_pci_dma_device_info_tbl[] = {
	{PCI_VENDOR_ID_NE, PCI_DEVICE_ID_NE},
	{PCI_VENDOR_ID_VF, PCI_DEVICE_ID_VF},
	{PCI_VENDOR_ID_INNO, PCI_DEVICE_ID_INNO},
	{PCI_VENDOR_ID_INNO_ALIAS, PCI_DEVICE_ID_G1_SOC_ALIAS},
	{PCI_VENDOR_ID_INNO, PCI_DEVICE_ID_G1_SOC},
	{PCI_VENDOR_ID_INNO, PCI_DEVICE_ID_G1_PAL},
	{PCI_VENDOR_ID_INNO, PCI_DEVICE_ID_G1_NE},
	{PCI_VENDOR_ID_INNO, PCI_DEVICE_ID_G0_SOC},
	{PCI_VENDOR_ID_INNO_ALIAS, PCI_DEVICE_ID_G0_SOC_ALIAS},
	{PCI_VENDOR_ID_INNO_ALIAS_GT, PCI_DEVICE_ID_G0_SOC_ALIAS},
	{PCI_VENDOR_ID_INNO, PCI_DEVICE_ID_G0_PAL},
	{PCI_VENDOR_ID_INNO, PCI_DEVICE_ID_G0_NE},
	{PCI_VENDOR_ID_INNO, PCI_DEVICE_ID_G1P_SOC},
	{PCI_VENDOR_ID_INNO, PCI_DEVICE_ID_G1P_PAL_VF},
	{PCI_VENDOR_ID_INNO_ALIAS, PCI_DEVICE_ID_G1P_SOC_ALIAS},
	{PCI_VENDOR_ID_INNO, PCI_DEVICE_ID_G1P_PAL},
	{PCI_VENDOR_ID_INNO, PCI_DEVICE_ID_G1P_NE},
	{PCI_VENDOR_ID_INNO, PCI_DEVICE_ID_G0M_SOC},
	{PCI_VENDOR_ID_INNO_ALIAS, PCI_DEVICE_ID_G0M_SOC_ALIAS},
	{PCI_VENDOR_ID_INNO, PCI_DEVICE_ID_G0M_PAL},
	{PCI_VENDOR_ID_INNO, PCI_DEVICE_ID_G0M_NE},
	{},
};

static struct pci_dev *mypdev = NULL;
static struct gpu_info_hwinfo *hwinfo;
static struct dev_rsrc *pdev_rsrc = NULL;
uint32_t mem_size = 0;

static struct role_target dma_get_role(int role_id)
{
	struct role_target role;

	role.vram_role = HAL_VRAM_ROLE_DMA;
	role.id = role_id;
	role.sub_id = 0;

	return role;
}

u64 __maybe_unused dma_alloc_vram(void *dev, int zone_id, uint64_t size)
{
        u64 dev_addr;
        struct role_target role;
        role = dma_get_role(zone_id);

        dev_addr = hal_vram_alloc(dev, &role, true, size);
        if (!dev_addr) {
                inno_error("hal vram alloc fail\n");
                return dev_addr;
        }
        return dev_addr;
}

void __maybe_unused dma_free_vram(void *dev, int zone_id, u64 dev_addr)
{
        struct role_target role;
        role = dma_get_role(zone_id);

        hal_vram_free(dev, &role, dev_addr);
}

static void init_data(void *address, uint64_t data, int size)
{
	int i = 0;
	void *addr = address;
	for (i = 0; i < size/8; i++) {
		WriteReg64(addr, data);
		addr += 8;
	}
}

#if 0
static void init_data_vram(void *address, uint64_t data, int size)
{
	int i = 0;
	void *addr = (void *)dev_paddr_to_cpu_paddr(&mypdev->dev, (uint64_t)address);
	addr = (void *)ioremap((uint64_t)addr, size);
	for (i = 0; i < size/8; i++) {
		WriteReg64(addr + i*8, data);
	}
	iounmap(addr);
}
#endif

static int data_check(void *src, void *dst, int size, unsigned int trans_type)
{
	int i = 0, ret = 0;
	void *src_data = src;
	void *dst_data = dst;

	if (DMA_DDR2DDR == trans_type || DMA_DDR2SYS == trans_type) {
		src_data = (void *)dev_paddr_to_cpu_paddr(&mypdev->dev, (uint64_t)src);
		src_data = (void *)ioremap((uint64_t)src_data, size);
	}

	if (DMA_DDR2DDR == trans_type || DMA_SYS2DDR == trans_type) {
		dst_data = (void *)dev_paddr_to_cpu_paddr(&mypdev->dev, (uint64_t)dst);
		dst_data = (void *)ioremap((uint64_t)dst_data, size);
	}

	for (i = 0; i < size/8; i++) {
		if(ReadReg64(src_data+i*8) != ReadReg64(dst_data+i*8))
		{
			pr_err("src_addr:%p dst_addr:%p\n",src_data+i*8, dst_data+i*8);
			pr_err("src:%#lx dst:%#lx\n",ReadReg64(src_data+i*8),ReadReg64(dst_data+i*8));
			ret = -1;
			break;
		}
	}

	if (DMA_DDR2DDR == trans_type || DMA_DDR2SYS == trans_type) {
		iounmap(src_data);
	}

	if (DMA_DDR2DDR == trans_type || DMA_SYS2DDR == trans_type) {
		iounmap(dst_data);
	}

	return ret;
}

static int data_check_smallbar(void *src, void *dst, int size, unsigned int trans_type)
{
	int i = 0, ret = 0;
	void *src_vis = src;
	void *dst_vis = dst;
	void *src_data = src;
	void *dst_data = dst;


	if(is_inv_mem)
	{
		if (DMA_DDR2SYS == trans_type)
		{
			src_vis = (void *)(sys_dev_addr + VIS_OFFSET);
			innodma_memcpy(&mypdev->dev, &src, &src_vis, &size, 1, GDDR2GDDR);
		}
		else if (DMA_SYS2DDR == trans_type)
		{
			dst_vis = (void *)(sys_dev_addr + VIS_OFFSET);
			innodma_memcpy(&mypdev->dev, &dst, &dst_vis, &size, 1, GDDR2GDDR);
		}
		else if (DMA_DDR2DDR == trans_type)
		{
			dst_vis = (void *)(sys_dev_addr + VIS_OFFSET);
			innodma_memcpy(&mypdev->dev, &dst, &dst_vis, &size, 1, GDDR2GDDR);
			src_vis = (void *)(sys_dev_addr + VIS_OFFSET + VIS_OFFSET);
			innodma_memcpy(&mypdev->dev, &src, &src_vis, &size, 1, GDDR2GDDR);
		}
	}

	if (DMA_DDR2DDR == trans_type || DMA_DDR2SYS == trans_type) {
		src_data = (void *)dev_paddr_to_cpu_paddr(&mypdev->dev, (uint64_t)src_vis);
		src_data = (void *)ioremap((uint64_t)src_data, size);
	}

	if (DMA_DDR2DDR == trans_type || DMA_SYS2DDR == trans_type) {
		dst_data = (void *)dev_paddr_to_cpu_paddr(&mypdev->dev, (uint64_t)dst_vis);
		dst_data = (void *)ioremap((uint64_t)dst_data, size);
	}

	for (i = 0; i < size/8; i++) {
		if(ReadReg64(src_data+i*8) != ReadReg64(dst_data+i*8))
		{
			pr_err("src_addr:%p dst_addr:%p\n",src_data+i*8, dst_data+i*8);
			pr_err("src:%#lx dst:%#lx\n",ReadReg64(src_data+i*8),ReadReg64(dst_data+i*8));
			ret = -1;
			break;
		}
	}

	if (DMA_DDR2DDR == trans_type || DMA_DDR2SYS == trans_type) {
		iounmap(src_data);
	}

	if (DMA_DDR2DDR == trans_type || DMA_SYS2DDR == trans_type) {
		iounmap(dst_data);
	}

	return ret;
}

static int get_dma_device(void)
{
	int i = 0;

	for (i = 0; i < ARRAY_SIZE(g_pci_dma_device_info_tbl); i++)
	{
		mypdev = pci_get_device(g_pci_dma_device_info_tbl[i].vendor_id, g_pci_dma_device_info_tbl[i].device_id, NULL);
		if (mypdev) {
			break;
		}
	}

	if (!mypdev)
	{
		pr_err("failed to get pci device!\n");
		return -1;
	}

	return 0;
}

static int src_dst_alloc(struct dma_memcpy_test_data *data, unsigned int trans_type)
{
	int i = 0;

	if(!data->dsts)
	{
		data->dsts=vmalloc(sizeof(void *)*data->cnt);
	}

	if(!data->dsts)
	{
		snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "data->dsts vmalloc error\n");
		return -1;
	}

	if(!data->srcs)
	{
		data->srcs=vmalloc(sizeof(void *)*data->cnt);
	}

	if(!data->srcs)
	{
		snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "data->srcs vmalloc error\n");
		return -1;
	}

	for(i = 0; i < data->cnt; i++)
	{
		if (DMA_DDR2DDR == trans_type || DMA_DDR2SYS == trans_type) {
			if(is_inv_mem) {
				if((INV_OFFSET + i*data->sizes[i]) > mem_size)
				{
					snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "vram offset is larger than vram size\n");
					goto e0;
				}
					data->srcs[i] = (void *)(sys_dev_addr + INV_OFFSET + i*data->sizes[i] + DIFF_OFFSET);
			}
			else
			{
				data->srcs[i] = (void *)dma_alloc_vram(&(mypdev->dev), 0, data->sizes[i]);
			}
			//init_data_vram(data->srcs[i], 0x5a5a5a5a5a5a5a5a, data->sizes[i]);
		}
		else
		{
			if (AXI_DMA == data->dma_type) {
			data->srcs[i] = kmalloc(data->sizes[i], GFP_KERNEL);
			}
			else {
			data->srcs[i] = vmalloc(data->sizes[i]);
			}
			init_data(data->srcs[i], 0x5a5a5a5a5a5a5a5a, data->sizes[i]);
		}

		if (data->srcs[i] == NULL) {
			snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "alloc source buffer failed\n");
			goto e0;
		}

		if (DMA_DDR2DDR ==  trans_type || DMA_SYS2DDR == trans_type) {
			if(is_inv_mem) {
				if((INV_OFFSET + (i+data->cnt) * data->sizes[i]) > mem_size)
				{
					snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "vram offset is larger than vram size\n");
					goto e0;
				}
				data->dsts[i] = (void *)(sys_dev_addr + INV_OFFSET + (i+data->cnt) * data->sizes[i]);
			}
			else {
				data->dsts[i] = (void *)dma_alloc_vram(&(mypdev->dev), 0, data->sizes[i]);
			}
			//init_data_vram(data->dsts[i], 0x0, data->sizes[i]);
		}
		else {
			if (AXI_DMA == data->dma_type) {
				data->dsts[i] = kmalloc(data->sizes[i], GFP_KERNEL);
			}
			else {
				data->dsts[i] = vmalloc(data->sizes[i]);
			}
			init_data(data->dsts[i], 0x0, data->sizes[i]);
		}

		if (data->dsts[i] == NULL) {
			snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "alloc destination buffer failed\n");
			goto e0;
		}
	}
	return 0;
e0:
	for(i = 0; i < data->cnt; i++)
	{
		if(data->srcs[i] == NULL)
			break;

		if (DMA_DDR2DDR == trans_type) {
			if(!is_inv_mem)
				dma_free_vram(&(mypdev->dev), 0, (u64)data->srcs[i]);
		}
		else
		{
			if (AXI_DMA == data->dma_type) {
				kfree(data->srcs[i]);
			} else
			{
				vfree(data->srcs[i]);
			}
		}
	}

	for(i = 0; i < data->cnt; i++)
	{
		if(data->dsts[i] == NULL)
			break;

		if (DMA_DDR2DDR ==  trans_type || DMA_SYS2DDR == trans_type) {
			if(!is_inv_mem)
				dma_free_vram(&mypdev->dev, 0, (u64)data->dsts[i]);

			if((DMA_DDR2DDR == trans_type) && is_inv_mem)
				dma_free_vram(&(mypdev->dev), 0, (u64)data->dsts[i]);
		}
		else {
			if (AXI_DMA == data->dma_type) {
				kfree(data->dsts[i]);
			}
			else {
				vfree(data->dsts[i]);
			}
		}
	}
	vfree(data->dsts);
	vfree(data->srcs);
	return -1;
}

static int src_dst_free(struct dma_memcpy_test_data *data, unsigned int trans_type)
{
	int i = 0;

	for(i = 0; i < data->cnt; i++)
	{
		if (DMA_DDR2DDR == trans_type || DMA_DDR2SYS == trans_type) {
			if(!is_inv_mem)
				dma_free_vram(&(mypdev->dev), 0, (u64)data->srcs[i]);
		}
		else
		{
			if (AXI_DMA == data->dma_type) {
				kfree(data->srcs[i]);
			} else
			{
				vfree(data->srcs[i]);
			}
		}

		if (DMA_DDR2DDR ==  trans_type || DMA_SYS2DDR == trans_type) {
			if(!is_inv_mem)
				dma_free_vram(&mypdev->dev, 0, (u64)data->dsts[i]);
		}
		else {
			if (AXI_DMA == data->dma_type) {
				kfree(data->dsts[i]);
			}
			else {
				vfree(data->dsts[i]);
			}
		}
	}

	vfree(data->dsts);
	vfree(data->srcs);
	kfree(data->sizes);
	return 0;
}

int gtest_dma_memcpy(struct dma_memcpy_test_data *data, size_t dir)
{
	int ret = 0;

	if(is_inv_mem && GDDR2GDDR != dir)
	{
		if(PCIE_DMA == data->dma_type)
		{
			ret = innodma_memcpy_for_smallbar_sg(&mypdev->dev, data->srcs, data->dsts, data->sizes, data->cnt, dir);
		}
		else if(AXI_DMA == data->dma_type)
		{
			ret = innodma_memcpy_for_smallbar(&mypdev->dev, data->srcs, data->dsts, data->sizes, data->cnt, dir);
		}
	}
	else
	{
		if(PCIE_DMA == data->dma_type)
		{
			ret = innodma_memcpy_sg(&mypdev->dev, data->srcs, data->dsts, data->sizes, data->cnt, dir);
		}
		else if(AXI_DMA == data->dma_type)
		{
			ret = innodma_memcpy(&mypdev->dev, data->srcs, data->dsts, data->sizes, data->cnt, dir);
		}
	}
	return ret;
}

int dma_open(struct inode *inode, struct file *flip)
{
	return 0;
}

ssize_t dma_read(struct file *filp, char __user *buf, size_t count, loff_t *fops)
{
	int ret = 0;

	ret = copy_to_user(buf, dma_buf, GTEST_DMA_CDEV_SIZE);
	if (ret != 0)
	{
		ret = -EFAULT;
	}

	return ret;
}

int dma_release(struct inode *inode, struct file *filp)
{
	// pr_info("pcie dma cdev release\n");
	return 0;
}

static long dma_ioctl(struct file * filp, unsigned int cmd, unsigned long args)
{
	int ret = 0, i = 0, j = 0;
	struct dma_memcpy_test_data *data;
	struct dma_memcpy_test_data user_data;
	int *user_sizes = NULL;

	if(args)
	{
		data = (struct dma_memcpy_test_data*)args;
		ret = copy_from_user(&user_data, (void*)args, sizeof(struct dma_memcpy_test_data));

		if(ret)
		{
			pr_err(" dma_ioctl: copy data from user space failed\n");
			snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "dma_ioctl: copy data from user space failed\n");
			goto e0;
		}

		user_sizes = kmalloc(sizeof(int*) * user_data.cnt, GFP_KERNEL);
		ret = copy_from_user(user_sizes, (void*)user_data.sizes, sizeof(int*) * user_data.cnt);

		if(ret)
		{
			pr_err(" dma_ioctl: copy data from user space failed\n");
			snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "dma_ioctl: copy data from user space failed\n");
			goto e0;
		}

		user_data.sizes = user_sizes;
		data = &user_data;
	}

	switch(cmd) {
	case DMA_SYS2DDR:
		ret = src_dst_alloc(data, cmd);
		if(ret)
		{
			goto e0;
		}

		for(i = 0; i < 10; i++)
		{
			ret = gtest_dma_memcpy(data, SYS2GDDR);
			if (ret)
			{
				snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "[FAIL] dma transfer: sys to ddr failed!\n");
				goto e1;
			}

			for(j = 0; j < data->cnt; j++)
			{
				if(is_inv_mem)
				{
					ret = data_check_smallbar(data->srcs[j], data->dsts[j], *data->sizes, cmd);
					if (ret)
					{
						snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "[FAIL] dma transfer: sys2ddr's data error!\n");
						goto e1;
					}
				}
				else
				{
					ret = data_check(data->srcs[j], data->dsts[j], *data->sizes, cmd);
					if (ret)
					{
						snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "[FAIL] dma transfer: sys2ddr's data error!\n");
						goto e1;
					}
				}
			}

		}
		snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "[SUCCESS] dma transfer: sys to ddr success!\n");
		break;
	case DMA_DDR2SYS:
		ret = src_dst_alloc(data, cmd);
		if(ret)
		{
			goto e0;
		}

		for(i = 0; i < 10; i++)
		{
			ret = gtest_dma_memcpy(data, GDDR2SYS);
			if (ret)
			{
				snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "[FAIL] dma transfer: ddr to sys failed!\n");
				goto e1;
			}

			for(j = 0; j < data->cnt; j++)
			{
				if(is_inv_mem)
				{
					ret = data_check_smallbar(data->srcs[j], data->dsts[j], *data->sizes, cmd);
					if (ret)
					{
						snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "[FAIL] dma transfer: sys2ddr's data error!\n");
						goto e1;
					}
				}
				else
				{
					ret = data_check(data->srcs[j], data->dsts[j], *data->sizes, cmd);
					if (ret)
					{
						snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "[FAIL] dma transfer: ddr2sys's data error!\n");
						goto e1;
					}
				}
			}
		}
		snprintf(dma_buf, GTEST_DMA_CDEV_SIZE,"[SUCCESS] dma transfer: ddr to sys success!\n");
		break;
	case DMA_DDR2DDR:
		ret = src_dst_alloc(data, cmd);
		if(ret)
		{
			goto e0;
		}

		for(i = 0; i < 10; i++)
		{
			ret = gtest_dma_memcpy(data, GDDR2GDDR);
			if (ret)
			{
				snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "[FAIL] dma transfer: ddr to ddr failed!\n");
				goto e1;
			}

			for(j = 0; j < data->cnt; j++)
			{
				if(is_inv_mem)
				{
					ret = data_check_smallbar(data->srcs[j], data->dsts[j], *data->sizes, cmd);
					if (ret)
					{
						snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "[FAIL] dma transfer: ddr2ddr's data error!\n");
						goto e1;
					}
				}
				else
				{
					ret = data_check(data->srcs[j], data->dsts[j], *data->sizes, cmd);
					if (ret)
					{
						snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "[FAIL] dma transfer: ddr2ddr's data error!\n");
						goto e1;
					}
				}
			}

		}
		snprintf(dma_buf, GTEST_DMA_CDEV_SIZE, "[SUCCESS] dma transfer: ddr to ddr success!\n");
		break;
	}

e1:
	src_dst_free(data, cmd);
e0:
	return 0;

}

static struct file_operations gtest_dma_filops =
{
	.owner = THIS_MODULE,
	.open = dma_open,
	.read = dma_read,
	.release = dma_release,
	.unlocked_ioctl = dma_ioctl,
	.compat_ioctl = dma_ioctl,
};

static int __init dma_cdev_init(void)
{
	int ret = 0;

	hwinfo = kmalloc(sizeof(struct gpu_info_hwinfo), GFP_KERNEL);
	if(!hwinfo)
	{
		inno_error("hwinfo alloc failed\n");
		return -1;
	}


	if (gtest_dma_major)
	{
		gtest_dma_cdev_id = MKDEV(gtest_dma_major, gtest_dma_minor);
		ret = register_chrdev_region(gtest_dma_cdev_id, 1, GTEST_DMA_TEST);
	}
	else
	{
		ret = alloc_chrdev_region(&gtest_dma_cdev_id, gtest_dma_minor, 1, GTEST_DMA_TEST);
		gtest_dma_major = MAJOR(gtest_dma_cdev_id);
		gtest_dma_minor = MINOR(gtest_dma_cdev_id);
	}

	if (ret < 0)
	{
		pr_err("failed to alloc gtest dma gtest_dma_cdev_id.\n");
		return ret;
	}

	cdev_init(&gtest_dma_cdev, &gtest_dma_filops);
	gtest_dma_cdev.owner = THIS_MODULE;
	gtest_dma_cdev.ops = &gtest_dma_filops;

	cdev_add(&gtest_dma_cdev, gtest_dma_cdev_id, 1);

	gtest_dma_test_class = class_create(THIS_MODULE, GTEST_DMA_TEST_CLASS);
	if (IS_ERR(gtest_dma_test_class))
	{
		pr_err("create class failed.\n");
		ret = -1;
		goto class_out;
	}

	gtest_dma_test_device = device_create(gtest_dma_test_class, NULL, gtest_dma_cdev_id, NULL, GTEST_DMA_TEST_DEV);
	if (IS_ERR(gtest_dma_test_device))
	{
		pr_err("create class failed.\n");
		ret = -1;
		goto device_out;
	}

	dma_buf = kmalloc(GTEST_DMA_CDEV_SIZE, GFP_KERNEL);

	ret = get_dma_device();
	if(ret)
	{
		return -1;
	}

	pdev_rsrc = inno_rsrc_devres_find(&mypdev->dev);

	if(!strcmp("g0m",hal_get_chip_name(&mypdev->dev)))
	{
		sys_dev_addr = G0M_VRAM_ADDR_START ;
	}

	is_inv_mem = (hal_has_inv_mem(&mypdev->dev));
	if(is_inv_mem)
	{
		hal_get_gpu_info_hwinfo(pdev_rsrc, hwinfo);
		mem_size = hwinfo->mem_width * hwinfo->mem_nums * 1024 * 1024 * 1024;
	}

	return 0;

device_out:
	device_destroy(gtest_dma_test_class, MKDEV(gtest_dma_major, 0));

class_out:
	class_destroy(gtest_dma_test_class);
	unregister_chrdev_region(gtest_dma_cdev_id, 1);
	return ret;
}

static void __exit dma_cdev_exit(void)
{
	pr_info("kfree %p\n", dma_buf);
	kfree((void *)dma_buf);
	kfree((void *)hwinfo);

	device_destroy(gtest_dma_test_class, MKDEV(gtest_dma_major, 0));
	class_destroy(gtest_dma_test_class);
	cdev_del(&gtest_dma_cdev);
	unregister_chrdev_region(gtest_dma_cdev_id, 1);
}


module_init(dma_cdev_init);
module_exit(dma_cdev_exit);

MODULE_LICENSE("Dual MIT/GPL");
